# General imports
import math
import periodictable
from itertools import zip_longest
import re

import digichem.log
from digichem.misc.base import dict_list_index
from digichem.exception.base import Result_unavailable_error, Digichem_exception
from digichem.result import Result_container
from digichem.result import Result_object
from digichem.result import Unmergeable_container_mixin
from digichem.file.prattle import Openprattle_converter

# Hidden import.


def get_chemical_group_mapping(rdkit_molecule):
    """
    Determine chemically equivalent atoms in this atom list.
    
    :return: A mapping between each group number and the atoms it contains.
    """
    from rdkit import Chem

    molecule = rdkit_molecule
    
    groupings = list(Chem.rdmolfiles.CanonicalRankAtoms(molecule, breakTies = False, includeChirality = False))
    
    atoms = list(molecule.GetAtoms())
    
    groups = {}
    for atom_index, atom in enumerate(atoms):
        # For most atoms, just use the assigned group number.
        group_num = groupings[atom_index]
        
        # If this atom is a hydrogen, and it has a single bond to a carbon, use that carbon's group number instead.
        if atom.GetSymbol() == "H":
            bonds = list(atom.GetBonds())
            if len(bonds) == 1 and bonds[0].GetOtherAtom(atom).GetSymbol() == "C":
                # This is an implicit H.
                other_atom = bonds[0].GetOtherAtom(atom)
                group_num = groupings[other_atom.GetIdx()]
        
        try:
            groups[(group_num, atom.GetSymbol())].append(atom_index +1)
        
        except KeyError:
            groups[(group_num, atom.GetSymbol())] = [atom_index +1]
            
    # We now need to re-assign group numbers to ensure they are:
    # 1) Sparse (no missing values, possible because of the hydrogens we removed).
    # 2) Consecutive in terms of the molecule's skeleton (so adjacent numbers are ideally placed
    # adjacent to each other).
    #
    # 2) is challenging to ensure, so for now we simply use the original atom position (because
    # molecules are normally written fairly consecutively).
    # TODO: Find a better way to ensure atom numbering is both deterministic and consecutive.
    
    order = []
    
    for atom_index, atom in enumerate(atoms):
        # Find the current group num of the group this atom is a part of.
        cur_group = dict_list_index(groups, atom_index +1)
        
        # If we've already processed this group, ignore.
        if cur_group[0] not in order:
            # Else, add it next.
            order.append(cur_group[0])
            
    new_groups = {(order.index(group_key[0]) +1, group_key[1]): group for group_key, group in groups.items()}
    
    return new_groups


class Nucleus():
    """
    A class that identifies a specific nucleus (having an element and isotope).
    """
    
    # A regex that can be used to identify a nucleus.
    regex = r'(\d+|\*)?([A-z][A-z]?)'
    
    def __init__(self, element):
        self.isotope = None
        
        if isinstance(element, type(self)):
            self.element = element.element
            self.isotope = element.isotope
        
        elif isinstance(element, list) or isinstance(element, tuple):
            # A list, first item is the element, second (if present) is the isotope).
            self.element = element[0]
            if len(element) > 1:
                self.isotope = element[1]
        
        elif isinstance(element, str):
            self.element, self.isotope = self.split_string(element)
        
        elif isinstance(element, int):
            self.element = periodictable.elements[element]
            
        else:
            raise TypeError("Unknown nucleus type '{}' {}".format(element, type(element)))
        
    @classmethod
    def split_string(self, string):
        """
        """
        match = re.search(self.regex, string)
        
        if match is None:
            raise ValueError("Unable to process nucleus string '{}'".format(string))
        
        match_groups = match.groups()
        isotope = int(match_groups[0]) if match_groups[0] != "*" else None
        
        if isinstance(match_groups[1], int):
            element = periodictable.elements[match_groups[1]]
            
        else:
            element = periodictable.elements.symbol(match_groups[1])
        
        return (element, isotope)
    
    def __str__(self):
        if self.isotope is not None:
            return "{}{}".format(self.isotope, self.element.symbol)
        
        else:
            return self.element.symbol
        

# Molecule_mixin is used for digichem.result.atom.Atom_list and digichem.input.digichem.Digichem_coords_ABC 
class Molecule_mixin():
    """
    Mixin for classes that represent molecules, compounds, or other molecular-like collections of atoms.
    
    Classes that inherit from this mixin must define 'element_dict' (as either an attribute or property).
    See the implementation in Atom_list for what this does.
    """
    
    # TODO: lots of opportunity to move stuff out of Atom_list and into this mixin.
    
    
    @property
    def formula(self):
        """
        Get a formula representation of this atom list.
        
        :return: The formula as a periodictable.formula object (which can be safely cast to string).
        """
        # A dictionary where each key is a type of atom (N, C, H etc) and the value is the number of that atom.
        atoms = self.element_dict
        # Build a string rep.
        form_string = ""
        for atom in atoms:
            form_string += "{}{}".format(atom, atoms[atom])

        # Get and return the formula object.
        return periodictable.formula(form_string)
        
    @property
    def formula_string(self):
        """
        Get a formula representation of this atom list as a string, including optional charge.
        """
        # Get the base formula
        #formula_string = str(self.formula)
        atoms = self.element_dict
        # Build a string rep.
        formula_string = ""
        for atom in atoms:
            formula_string += "{}{}".format(atom, atoms[atom])
        
        # Add charge, if we have one.
        if self.charge == 1:
            formula_string += " +"
        elif self.charge == -1:
            formula_string += " -"
        elif self.charge != 0:
            formula_string += " {}{}".format(abs(self.charge), "-" if self.charge < 0 else "+")
            
        return formula_string
        
    @property
    def molar_mass(self):
        """
        The molar mass of the molecule (takes into account different isotopes and relative isotope abundances, unlike the mass attribute).
        
        :return: The mass (in Daltons / gmol-1).
        """
        return self.formula.mass
    

class Atom_list(Result_container, Unmergeable_container_mixin, Molecule_mixin):
    """
    Class for representing a group of atoms.
    """
    
    # A warning issued when attempting to merge non-equivalent atom lists.
    MERGE_WARNING = "Attempting to merge lists of atoms that are not identical; non-equivalent atoms will be ignored"
    
    def __init__(self, *args, charge = None, assign_groups = True, **kwargs):
        """
        :param charge: The electronic charge of the system.
        :param assign_groups: If True (the default), assign groupings to all the atoms of this list. Call assign_groups() to re-calculate these groupings whenever atoms are added/removed.
        """
        super().__init__(*args, **kwargs)
        self.charge = charge if charge is not None else 0
        self._groups = {}
        
        if assign_groups:
            self.groups
        
    @property
    def mass(self):
        """
        The total mass of all the atoms in this set.
        
        :return: The mass (in Daltons).
        """
        try:
            return sum([atom.mass for atom in self])
        except TypeError:
            # Exact mass not available.
            raise Result_unavailable_error("Exact mass") from None
        
    @property
    def element_dict(self):
        """
        Get a dictionary where each key is one of the elements in the atom list (C, H, N etc) and the value is the number of that element that appears in the atom list.
        
        :return: The element dictionary.
        """
        atoms = {}
        for atom in self:
            # Try and increment the count of the atom.
            try:
                atoms[atom.element.symbol] += 1
            except KeyError:
                # Add the new atom.
                atoms[atom.element.symbol] = 1
        return atoms
    
    @property
    def groups(self):
        """
        Get a list of the groups of atoms in this list.
        
        Atom groups combine chemically equivalent atom positions into a single group.
        """
        if self._groups == {} and len(self) > 0:
            self.assign_groups()
            
        return self._groups
    
    def find(self, criteria = None, *, label = None, index = None):
        """
        Find an atom that matches a given criteria
        
        :raises ValueError: If the requested atom could not be found.
        :param criteria: Automatically determine which criteria to search by.
        :param label: The label (a string, such as C1), to find.
        :param index: The index (an int or string that looks like an int) of the atom
        :return: The requested atom.
        """
        if criteria is not None:
            if criteria.isdigit() or isinstance(criteria, int):
                index = int(criteria)
            else:
                label = criteria
        
        if index:
            # Fetch by index.
            try:
                return self[int(index)]
            
            except IndexError:
                raise Result_unavailable_error("Atom", "could not find atom with index {}".format(index)) from None
            
        else:
            # Search by label.
            try:
                return [atom for atom in self if atom.label == label][0]
            
            except IndexError:
                raise Result_unavailable_error("Atom", "could not find atom with label {}".format(label)) from None
    
    def assign_groups(self):
        """
        Assign groupings to all the atoms of this system.
        
        Atom groups combine chemically equivalent atom positions into a single group.
        """
        group_mappings = get_chemical_group_mapping(self.to_rdkit_molecule())            
        self._groups = {group_id: Atom_group(group_id[0], [self[atom_index -1] for atom_index in atom_indices]) for group_id, atom_indices in group_mappings.items()}
        
        for atom in self:
            group = [group for group in self.groups.values() if atom in group.atoms][0]
            
            atom.group = group
    
    @property
    def smiles(self):
        """
        Get this geometry in (canonical) SMILES format.
        """
        try:
            return self._smiles
        
        except AttributeError:
            # Cache miss, go do some work.

            from rdkit.Chem import MolToSmiles
            from rdkit.Chem.rdmolops import RemoveHs
            
            # TODO: Find some other way of generating SMILES.

            mol = self.to_rdkit_molecule()
            try:
                # TODO: rdkit is unreliable, this method can fail for lots of reasons...
                mol = RemoveHs(mol)
            except Exception:
                pass
                
            self._smiles = MolToSmiles(mol)
            return self._smiles

            # # TODO: Handle cases where obabel isn't available
            # conv = Openprattle_converter.get_cls("xyz")(input_file = self.to_xyz(), input_file_type = "xyz")
            # # Cache the result in case we need it again.
            # self._smiles = conv.convert("can").strip()
            # return self._smiles
        
    @property
    def X_length(self):
        return self.get_axis_length(0)
    
    @property
    def Y_length(self):
        return self.get_axis_length(1)
    
    @property
    def Z_length(self):
        return self.get_axis_length(2)
    
    def get_axis_length(self, axis):
        """
        Calculate the length of an axis, defined as the distance required in that axis to contain all the atoms of the set.
        
        :param axis: The axis to calculate for as an integer (0: X-axis, 1: Y-axis, 2: Z-axis).
        :return The length (in angstroms).
        """
        if not 0 <= axis <= 2:
            # Axis is invalid.
            raise ValueError("Axis '{}' is out of bounds. Possible  values are 0 (X), 1 (Y) or 2 (Z)")
        
        # First sort our list of atoms in terms of x, y or z coord.
        sorted_atoms = sorted(self, key = lambda atom: atom.coords[axis])
                
        # Now the axis length is simply the difference between the greatest and the smallest.
        try:
            return sorted_atoms[-1].coords[axis] - sorted_atoms[0].coords[axis]
        
        except IndexError:
            if len(sorted_atoms) == 0:
                # There are not atoms.
                return 0.0
            
            else:
                raise
    
    def get_linear_ratio(self):
        """
        Get the linear ratio of the molecule.
        
        The linear ratio is defined as 1 - (Y_length / X_length).
        
        :return: The ratio, from 0 (non-linear) to 1 (linear).
        """
        try:
            return 1- (self.Y_length / self.X_length)
        
        except (FloatingPointError, ZeroDivisionError):
            return 0
    
    def get_planar_ratio(self):
        """
        Get the planar ratio of the molecule.
        
        The planar ratio is defined as 1 - (Z_length / Y_length).
        
        :return: The ratio, from 0 (non-planar) to 1 (planar).
        """
        try:
            return 1- (self.Z_length / self.Y_length)
        
        except (FloatingPointError, ZeroDivisionError):
            return 0
    
    def get_X_axis_angle(self, start_coord, end_coord):
        """
        Get the angle between a line and the X axis.
        
        :param start_coord: A (X, Y, Z) tuple of coordinates of the start of the line.
        :param end_coord: A (X, Y, Z) tuple of coordinates of the end of the line.
        :return: The angle (in radians). 
        """
        return self.get_theta(math.sqrt( (end_coord[2] - start_coord[2])**2 + (end_coord[1] - start_coord[1])**2 ), end_coord[0] - start_coord[0])
    
    def get_XY_plane_angle(self, start_coord, end_coord):
        """
        Get the angle between a line and the XY plane.
        
        :param start_coord: A (X, Y, Z) tuple of coordinates of the start of the line.
        :param end_coord: A (X, Y, Z) tuple of coordinates of the end of the line.
        :return: The angle (in radians). 
        """
        # The 'secondary' axis is the opposite side of our triangle.
        secondary_axis = end_coord[2] - start_coord[2]
        # The 'primary' axis is the adjacent side of our triangle, which we can get with pythagoras.
        primary_axis = math.sqrt( (end_coord[0] - start_coord[0])**2 + (end_coord[1] - start_coord[1])**2 )
        return self.get_theta(secondary_axis, primary_axis)
    
    @classmethod
    def from_parser(self, parser):
        """
        Get an Atom_list object from an output file parser.
        
        :param parser: An output file parser.
        :param charge: Charge of the system.
        :return: A list of TDM objects.
        """
        return self(Atom.list_from_parser(parser), charge = parser.results.metadata.charge)
    
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get an instance of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return self(Atom.list_from_dump(data['values'], result_set, options), charge = data['charge'])
    
    @classmethod
    def from_coords(self, coords):
        """
        Get an instance of this class from a Digichem input coordinates object.
        
        :param coords: Digichem input coords.
        """
        return self(Atom.list_from_coords(coords), charge = coords.charge)
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        try:
            smiles = self.smiles
        
        except Exception:
            smiles = None
        
        dump_dict = {
            "formula": self.formula_string,
            "charge": self.charge,
            "smiles": smiles,
            "exact_mass": {
                "value": float(self.mass) if self.safe_get("mass") is not None else None,
                "units": "g mol^-1" 
            },
            "molar_mass": {
                "value": self.molar_mass,
                "units": "g mol^-1",
                },
            "num_atoms": len(self),
            "x-extension": {
                "value": float(self.X_length),
                "units": "Å"
            },
            "y-extension": {
                "value": float(self.Y_length),
                "units": "Å"
            },
            "z-extension": {
                "value": float(self.Z_length),
                "units": "Å"
            },
            "linearity_ratio": float(self.get_linear_ratio()),
            "planarity_ratio": float(self.get_planar_ratio()),
            "values": super()._dump_(digichem_options, all),
        }
        return dump_dict
    
    @classmethod
    def merge(self, *multiple_lists, charge):
        """
        Merge multiple lists of atoms into a single list.
         
        Note that it does not make logical sense to combine different list of atoms into one; hence the method only ensures that all given lists (which are not empty) are the same and then returns the first (non empty) given.
        If the atom lists are not equivalent, a warning will be issued.
        """
        return super().merge(*multiple_lists, charge = charge)
    
    def to_xyz(self):
        """
        Convert this list of atoms to xyz format.
        """
        # First, the number of atoms.
        xyz = "{}\n\n".format(len(self))
        
        # Then coordinates.
        # No effort is made here to truncate coordinates to a certain precision.
        for atom in self:
            xyz += "{}    {:f}    {:f}    {:f}\n".format(atom.element.symbol, float(atom.coords[0]), float(atom.coords[1]), float(atom.coords[2]))
        
        return xyz
    
    def to_mol(self):
        """
        Convert this list of atoms to mol format (useful for reading with rdkit).
        """
        return Openprattle_converter(input_file_buffer = self.to_xyz(), input_file_path = "internal atoms object", input_file_type = "xyz").convert("mol", charge = self.charge)
    
    def to_rdkit_molecule(self):
        """
        Convert this list of atoms to an rdkit molecule object.
        """
        from rdkit import Chem
        from rdkit.Chem import rdDetermineBonds
        
        # RDKit has a lot of problems loading molecule information from other formats.
        # Loading from Mol specifies atom and bonding types, but rdkit is very fragile when it comes to
        # which bonds it considers allowed (tetravelent nitrogen is not allowed for example). This can
        # be 'fixed'/ignored by specifying sanitize = False, but then lots of other rdkit functions don't
        # work properly, and rdkit does not detect symmetry correctly.
        # Loading from SMILES is a bad idea because hydrogens are lost.
        # Loading from xyz is much more robust and maintains Hs, but no bond information is read
        # (because there isn't any). RDkit can try and determine bonding itself with DetermineBonds(),
        # but this function is also fragile.
        
        # NOTE: rdkit can fail silently and return None, best to check.
#        mol = Chem.MolFromMolBlock(self.to_mol(), removeHs = False, sanitize = False)
        
        
#         parse_settings = Chem.rdmolfiles.SmilesParserParams()
#         parse_settings.removeHs = False
#         mol = Chem.MolFromSmiles(self.smiles, parse_settings)
#         
        mol = Chem.MolFromXYZBlock(self.to_xyz())
        if mol is None:
            raise Exception("Failed to parse coordinates with rdkit")
        
        mol.UpdatePropertyCache()
        rdDetermineBonds.DetermineConnectivity(mol, charge = self.charge)
        try:
            rdDetermineBonds.DetermineBonds(mol, charge = self.charge)
        
        except Exception:
            # This function is not implemented for some atoms (eg, Se).
            digichem.log.get_logger().warning(
                "Unable to determine bond ordering for molecule; all bonds will be represented as single bonds only".format(self.formula_string)
                , exc_info = True
            )
        
        return mol


class Atom_ABC(Result_object):
    """
    ABC for atom-like result objects.
    """
    
    @property
    def label(self):
        return "{}{}".format(self.element, self.index)
    
    def __hash__(self):
        return hash(self.index)


class Atom_group(Atom_ABC):
    """
    A class that represents a group of atoms that are chemically equivalent.
    """
    
    def __init__(self, index, atoms):
        self.index = index
        self.atoms = atoms
        # Check all elements are the same.
        self.element
        
    @property
    def element(self):
        elements = list(set(atom.element for atom in self.atoms))
        
        if len(elements) > 1:
            raise Digichem_exception("Multiple element types found in atom group '{}'".format(elements))
        
        return elements[0]
    
    @property
    def id(self):
        return (self.index, self.element.symbol)
    
    @property
    def label(self):
        return "{}{}".format(self.element, self.index)
    
    def __str__(self):
        """
        Stringify this atom group.
        """
        return "G:{}".format(self.label)
        
    def __eq__(self, other):
        """
        Is this atom group equal to another?
        """
        return self.atoms == other.atoms
    
    def __hash__(self):
        return super().__hash__()


class Atom(Atom_ABC):
    """
    Class that represents an atom.
    """
    
    def __init__(self, index, atomic_number, coords, mass = None):
        """
        Construct an Atom class.
        
        :param index: The numerical index of this atom, starting from 1.
        :param atomic_number: The atomic/proton number of the atom. Conventional wisdom suggests this has to be an integer, but we make no check here.
        :param mass: The mass of the atom (in Daltons). This isn't always available for some reason (eg, Gaussian Freq calculations), but when it is it identifies the isotope of the given atom.
        :param coords: The coords of the atom, as an (x, y, z) tuple.        
        """
        # Just save each of our attributes.
        self.index = index
        self.mass = mass
        self.coords = coords
        # Get our element class.
        self.element = periodictable.elements[atomic_number]
        # Atom groups are assigned by Atom_list objects.
        self.group = None
        
    def __str__(self):
        """
        Stringify this atom.
        """
        return "'{}' at '{}'".format(self.label, self.coords)
        
    def __eq__(self, other):
        """
        Is this atom equal to another?
        """
        # Atoms are considered equal if they are the same element in the same position.
        #return self.element == other.element and self.coords == other.coords
        return self.element == other.element and self.index == other.index
    
    def __hash__(self):
        return super().__hash__()
            
    def distance(self, foreign_atom):
        """
        Get the distance between this atom and another atom.
        
        :return: The distance. The units depend on the units of the atoms' coordinates. If the two atoms have coordinates of different units, then you will obviously get bizarre results.
        """
        return math.sqrt( (self.coords[0] - foreign_atom.coords[0])**2 + (self.coords[1] - foreign_atom.coords[1])**2 + (self.coords[2] - foreign_atom.coords[2])**2)
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        dump_dict = {
            "index": self.index,
            "element": self.element.number,
            "label": self.label,
            "group": self.group.label if self.group else None,
            "coords": {
                "x": {
                    "value": float(self.coords[0]),
                    "units": "Å", 
                },
                "y": {
                    "value": float(self.coords[1]),
                    "units": "Å", 
                },
                "z": {
                    "value": float(self.coords[2]),
                    "units": "Å", 
                }
            },
            "mass": {
                "value": float(self.mass) if self.mass is not None else None,
                "units": "g mol^-1"
            }
        }
        return dump_dict
        
    @classmethod
    def list_from_dump(self, data, result_set, options):
        """
        Get a list of instances of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return [self(atom_dict.get('index', index +1), atom_dict['element'], (float(atom_dict['coords']['x']['value']), float(atom_dict['coords']['y']['value']), float(atom_dict['coords']['z']['value'])), atom_dict['mass']['value']) for index, atom_dict in enumerate(data)]
    
    @classmethod
    def list_from_parser(self, parser):
        """
        Get a list of Atom objects from an output file parser.
        
        :param parser: An output file parser.
        :result: A list of Atom objects. An empty list is returned if no atom data is available.
        """
        # First pack our data together to make is easier to loop through.
        try:
            atomnos = parser.data.atomnos
            # Atom coords contains a list for each iteration, we only want the last.
            atomcoords = parser.data.atomcoords[-1]
            atommasses = getattr(parser.data, 'atommasses', [])
            
            # Atommasses sometimes is longer than atomcoords or atomnos.
            # This might be a cclib bug, it is not clear.
            if len(atommasses) > len(atomnos):
                # Take the last values only.
                atommasses = atommasses[-len(atomnos):]
            
        except AttributeError:
            # No atom data available.
            return []
        
        # Zip.
        zip_data = zip_longest(atomnos, atomcoords, atommasses, fillvalue = None)
        
        # Loop through and rebuild our objects.
        return [self(index+1, atomic_number, (float(coords[0]), float(coords[1]), float(coords[2])), mass) for index, (atomic_number, coords, mass) in enumerate(zip_data)]
    
    @classmethod
    def list_from_coords(self, coords):
        """
        Get a list of Atom objects from a Digichem input coordinates object.
        
        :param coords: Digichem input coords.
        :result: A list of Atom objects. An empty list is returned if no atom data is available.
        """
        return [self(index+1, getattr(periodictable.elements, atom['atom']).number, (atom["x"], atom["y"], atom["z"])) for index, atom in enumerate(coords.atoms)]



import math

import digichem.config


class Angle():
    """
    A class for representing angles (either in radians or degrees).
    """
    
#     # The default units to print angles in.
#     _default_angle_units = "deg"
#     
#     @classmethod
#     def set_default_angle_units(self, angle_units):
#         """
#         Change the default angle units used by subsequent objects.
#         """
#         if angle_units != "rad" and angle_units != "deg":
#             self._raise_angle_unit_error(angle_units)
#         else:
#             self._default_angle_units = angle_units

    @property
    def _default_angle_units(self):
        return  digichem.config.get_config()['angle_units']
            
    
    def __init__(self, angle, angle_units = None, output_units = "def"):
        """
        Constructor for angle objects.
        
        :param angle: The numerical value of the angle. Note that this will be stored internally in radians under the 'radians' attribute, and as degrees under 'degrees'.
        :param angle_units: The units of the angle, either "rad" for radians or "deg" for degrees. If angle is an Angle object, angle_units is ignored and determined automatically. Otherwise and if angle_units is None, "rad" is assumed as the default.
        :param output_units: The units to use when accessing the angle, either "def" to use the class default 'default_angle_units', "rad" for radians or "deg" for degrees. This can be changed post init by setting the 'units' property.
        """
        if isinstance(angle, Angle):
            angle_units = angle.units
            angle = float(angle)
        
        if angle_units is None:
            angle_units = "rad"
        
        # Check our units are valid and save our angle.
        if angle_units == "rad":
            self._angle = angle
        elif angle_units == "deg":
            self._angle = self.deg_to_rad(angle)
        else:
            # Get upset.
            self._raise_angle_unit_error(angle_units)
        
        # Save our output units (this is type checked for us).
        if output_units == "def":
            # Use our default (which can be set at the module level).
            self.units = self._default_angle_units
        else:
            self.units = output_units
            
    @classmethod
    def _raise_angle_unit_error(self, angle_unit):
        """
        Convenience function that raises an exception.
        """
        raise ValueError('\'{}\' is not a valid angle unit. Accepted values are "rad" or "deg"'.format(angle_unit))
            
    @property
    def radians(self):
        """
        The angle in radians.
        """
        return self._angle
    
    @property
    def degrees(self):
        """
        The angle in degrees.
        """
        return self.rad_to_deg(self._angle)
    
    @property
    def angle(self):
        if self._units == "rad":
            return self._angle
        elif self._units == "deg":
            return self.rad_to_deg(self._angle)
        else:
            self._raise_angle_unit_error(self._units)
        
    @property
    def units(self):
        """
        The units of the angle, either "deg" for degrees or "rad" for radians.
        """
        return self._units
    
    @units.setter
    def units(self, value):
        if value == "rad" or value == "deg":
            self._units = value
        else:
            self._raise_angle_unit_error(self._units)
    
    @property
    def pretty_units(self):
        """
        The units of the angle, currently either "°" for degrees or "rad" for radians.
        
        Note that the units returned by this method may change without warning so they are not suitable for type checking, use 'units' for actually checking the type of angle.
        """
        return self.units_to_pretty_units(self.units)
        
    @classmethod
    def units_to_pretty_units(self, units):
        """
        Convert a string describing angle units (either 'deg' or 'rad') to an appropriate symbol, currently either "°" for degrees or "rad" for radians.
        
        :param units: The units as a string.
        :return: An appropriate unit symbol as a string.
        """
        if units == "deg":
            return "°"
        elif units == "rad":
            return "rad"
        else:
            return ""
        
    
    def __float__(self):
        """
        Get the numerical value of this angle, according to our current 'units'.
        
        Use either the 'radians' or 'degrees' properties to use a specific unit.
        """
        return self.angle
    
    def __str__(self):
        """
        String representation of this angle, currently in the format: "angle" "units".
        """
        return "{} {}".format(self.angle, self.pretty_units)
            
            
    @classmethod
    def deg_to_rad(self, deg):
        return math.radians(deg)
    
    @classmethod
    def rad_to_deg(self, rad):
        return math.degrees(rad)
            
            
        


import itertools

from digichem.exception.base import Result_unavailable_error
from digichem.result.dipole_moment import Electric_dipole_moment_mixin, Dipole_moment_ABC, Magnetic_dipole_moment_mixin
from digichem.result.base import Result_object


class Transition_dipole_moment_ABC(Dipole_moment_ABC):
    """
    ABC for electric and magnetic transition dipole moments.
    """
    
    def __init__(self, state_level, *args, **kwargs):
        """
        Electric_transition_dipole_moment constructor.
        
        :param state: The excited state level this dipole transitions to.
        """
        super().__init__(*args, **kwargs)
        
        # An int describing the excited state we belong to. This is always available.
        self.state_level = state_level
        
        # The excited state object that we belong to. This may be None. We can't set this here because then both the Excited_state and Transition_dipole_moment constructors would depend on each other.
        self.excited_state = None
        
        # Save a name describing which dipole we are (permanent vs transition etc).
        self.dipole_type = "transition"
        
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get a list of instances of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return self(
            data['state_level'],
            (data['origin']['x']['value'], data['origin']['y']['value'], data['origin']['z']['value']),
            (data['vector']['x']['value'], data['vector']['y']['value'], data['vector']['z']['value']),
            atoms = result_set.atoms
            )
        
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        data = {'state_level': self.state_level}
        data.update(super()._dump_(digichem_options, all))
        return data
    
    @property
    def name(self):
        """
        Name that describes this dipole moment.
        """
        return "{} TDM".format(self.excited_state.state_symbol)
    
    def set_excited_state(self, excited_state):
        """
        Set the excited state object that we belong to.
        """
        # Not sure what we should do if the given excited_state has a different level to what we expect.
        if excited_state.level != self.state_level:
            pass
        self.excited_state = excited_state
        
        
class Electric_transition_dipole_moment(Transition_dipole_moment_ABC, Electric_dipole_moment_mixin):
    """
    Class that represents an electric transition dipole moment.
    """
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.electromagnetic_type = "electric"
        
    @classmethod
    def list_from_parser(self, parser):
        """
        Get a list of TDMs from an output file parser.
        
        :param parser: An output file parser.
        :return: A list of TDM objects. An empty list will be returned if no TDM data is available.
        """
        try:
            return [
                self(state_level = index +1,
                origin_coords = (0,0,0),
                vector_coords = (parser.au_to_debye(tdm[0]), parser.au_to_debye(tdm[1]), parser.au_to_debye(tdm[2])),
                atoms = parser.results.atoms) for index, tdm in enumerate(parser.data.etdips)]
        except AttributeError:
            return []


class Magnetic_transition_dipole_moment(Transition_dipole_moment_ABC, Magnetic_dipole_moment_mixin):
    """
    Class that represents a magnetic transition dipole moment.
    
    Unlike electric dipole moments, the units here are in a.u. (0.5 bohr magnetons).
    """
        
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.electromagnetic_type = "magnetic"
    
    @classmethod
    def list_from_parser(self, parser):
        """
        Get a list of TDMs from an output file parser.
        
        :param parser: An output file parser.
        :return: A list of TDM objects. An empty list will be returned if no TDM data is available.
        """
        try:
            return [
                self(state_level = index +1,
                origin_coords = (0,0,0),
                vector_coords = tdm,
                atoms = parser.results.atoms) for index, tdm in enumerate(parser.data.etmagdips)]
        except AttributeError:
            return []


class Transition_dipole_moment(Result_object):
    """
    A compound class that represents both the electric and magnetic components of a transition dipole moment.
    
    This class can also be used in most places where an electric TDM is expected, as it will pass references to TEDM attributes to the actual TEDM class.
    If the object has only a TMDM and no TEDM, references will instead be passed to the TMDM.
    """
    
    def __init__(self, electric = None, magnetic = None):
        """
        Constructor for TDM.
        
        Either or both of the substituent TDM can be None.
        """
        self.electric = electric
        self.magnetic = magnetic
        self.electromagnetic_type = "electromagnetic"
    
    def set_excited_state(self, excited_state):
        """
        Set the excited state object that we belong to.
        """
        if self.electric is not None:
            self.electric.set_excited_state(excited_state)
        
        if self.magnetic is not None:
            self.magnetic.set_excited_state(excited_state)
        
    @classmethod
    def from_parser(self, parser):
        return self.list_from_parser(parser)
    
    @classmethod
    def list_from_parser(self, parser):
        """
        Get a list of Transition_dipole_moment from an output file parser.
        
        :param parser: An output file parser.
        :return: A list of Transition_dipole_moment objects. An empty list will be returned if no TDM data is available.
        """
        electric = Electric_transition_dipole_moment.list_from_parser(parser)
        magnetic = Magnetic_transition_dipole_moment.list_from_parser(parser)
        
        return [self(electric_part, magnetic_part) for electric_part, magnetic_part in itertools.zip_longest(electric, magnetic)]
        
    def __getattr__(self, name):
        # This check prevents infinite recursion when pickling, and is probably a sensible check anyway...
        if name != "electric" and name != "magnetic":
            try:
                if self.electric is not None:
                    return getattr(self.electric, name)
                else:
                    return getattr(self.magnetic, name)
            except AttributeError as e:
                raise AttributeError(name) from e
        else:
            raise AttributeError(name)
        
    @classmethod
    def list_from_dump(self, data, result_set, options):
        """
        Get a list of instances of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        # This constructor for tdms is a bit odd in that it does not work anything like the other from_dump() methods.
        # In the dumped format, tdms are not stored as their own list (but rather as a sub dict of the relevant excited state),
        # but in order to avoid recursive imports, a list of tdms does need to be created when excited states are loaded again.
        # This being the case, 'data' here is actually a list of excited states, not tdms...
        return [self.from_dump(excited_state['tdm'], result_set, options) for excited_state in data]
        
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get a list of instances of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        if data is None:
            return None
    
        # Build the individual parts.
        if data['electric'] is not None:
            electric = Electric_transition_dipole_moment.from_dump(data['electric'], result_set, options)
        
        else:
            electric = None
        
        if data['magnetic'] is not None:
            magnetic = Magnetic_transition_dipole_moment.from_dump(data['magnetic'], result_set, options)
        
        else:
            magnetic = None
        
        return self(electric = electric, magnetic = magnetic)
        
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        return {
            "electric": self.electric.dump(digichem_options, all) if self.electric is not None else None,
            "magnetic": self.magnetic.dump(digichem_options, all) if self.magnetic is not None else None,
            "angle": {
                "value": float(self.angle().angle) if self.electric is not None and self.magnetic is not None else None,
                "units": self.angle().units if self.electric is not None and self.magnetic is not None else None,
            },
            "cos_angle": float(self.cos_angle()) if self.electric is not None and self.magnetic is not None else None,
            "dissymmetry_factor": float(self.g_value) if self.safe_get('g_value') is not None else None
        }
            
    
    def angle(self, cgs = True):
        """
        Return the angle between the electric and magnetic components of this transition dipole moment.
        """
        return self.electric.angle(self.magnetic, cgs)
    
    def cos_angle(self, cgs = True):
        """
        Return the cos of the angle between the electric and magnetic components of this transition dipole moment.
        """
        return self.electric.cos_angle(self.magnetic, cgs)
    
    @property
    def g_value(self):
        """
        Return the 'g value'; the dissymmerty factor of this transition dipole moment.
        
        This calculation is based on J. Phys. Chem. Lett. 2021, 12, 1, 686–695.
        """
        try:
            return (4 * self.electric.gaussian_cgs * self.magnetic.gaussian_cgs * self.cos_angle(True)) / (self.electric.gaussian_cgs **2 + self.magnetic.gaussian_cgs **2)
        
        except (FloatingPointError, ZeroDivisionError):
            return 0
        
        except AttributeError:
            raise Result_unavailable_error("g_value", "transition electric or transition magnetic dipole moment is not available") from None




"""Tests for calculated results"""

import pytest
from pathlib import Path
import shutil

from digichem.test.util import data_directory, digichem_options
from digichem.parse.util import parse_calculation


@pytest.fixture(scope="module")
def formaldehyde_SOC_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Formaldehyde/Gaussian 09 Excited States TD-DFT 4 Singlets 4 Triplets B3LYP Gas Phase TZVP.tar.gz"), options = digichem_options)

@pytest.mark.skipif(not shutil.which("rwfdump"), reason="No rwfdump available")
@pytest.mark.parametrize("result_set, singlet, triplet, SOC", [
    # These SOC values are taken from the original paper J. Chem. Theory Comput. 2017, 13, 515−524 for PySOC,
    # the program Silioc uses to calculate SOC. We are limited in terms of accuracy by the number of decimals
    # (0) that they report...
        (pytest.lazy_fixture("formaldehyde_SOC_result"), "S(1)", "T(1)", 0.0),
        (pytest.lazy_fixture("formaldehyde_SOC_result"), "S(1)", "T(2)", 45.0)
    ])
def test_soc(result_set, singlet, triplet, SOC):
    """Test calculation of spin-orbit coupling values."""
    
    # Check we have SOC between each singlet state (including the ground) and each triplet state.
    assert len(result_set.soc) == (result_set.excited_states.num_singlets +1) * result_set.excited_states.num_triplets
    
    # Check some SOC values.
    assert result_set.soc.between(singlet, triplet).wavenumbers == pytest.approx(SOC, abs=1e-0)


# Code for processing results from calculations and generating reports.

# General imports.
import itertools
import warnings

from digichem.exception import Result_unavailable_error

from digichem.result import Result_object
import digichem.result.excited_state
from digichem.result.emission import Emissions


class Result_set(Result_object):
    """
    Class that represents a collection of results from a calculation.
    
    This class is a bit heavy and might not be around for long...
    """
    
    def __init__(self, **attributes):
        """Constructor for Result_set objects."""
        super().__init__()
        
        # The ID uniquely identifies this calculation result.
        # In most cases, it is generated from the combined chekcsum of the log files from which the results was parsed.
        # If this result was not read from log files (but was read from a database or from a .si calc result or similar),
        # then the id will simply be the same as was written previously.
        #
        # Note that the underscore here does not imply that this attribute is 'hidden'. Instead, this is used for 
        # consistency with mongo-like DBs, where the ID field is also typically named _id.
        #self._id = attributes.pop('database_id', None)
        self._id = attributes.pop('_id', None)
        
        self.results = (self,)
        self.emission = attributes.pop('emission', Emissions())

        # Any ancillary data.
        self._aux = attributes.pop('aux', {})
        
        for attr_name, attribute in attributes.items():
            setattr(self, attr_name, attribute)
        
    @property
    def alignment(self):
        warnings.warn("alignment is deprecated, use atoms instead", DeprecationWarning)
        return self.atoms

    @property
    def CC_energies(self):
        warnings.warn("CC_energies is deprecated, use energies.cc instead", DeprecationWarning)
        return self.energies.cc
    
    @CC_energies.setter
    def CC_energies(self, value):
        warnings.warn("CC_energies is deprecated, use energies.cc instead", DeprecationWarning)
        self.energies.cc = value
    
    @property
    def MP_energies(self):
        warnings.warn("MP_energies is deprecated, use energies.mp instead", DeprecationWarning)
        return self.energies.mp
    
    @MP_energies.setter
    def MP_energies(self, value):
        warnings.warn("MP_energies is deprecated, use energies.mp instead", DeprecationWarning)
        self.energies.mp = value
    
    @property
    def SCF_energies(self):
        warnings.warn("SCF_energies is deprecated, use energies.scf instead", DeprecationWarning)
        return self.energies.scf
    
    @SCF_energies.setter
    def SCF_energies(self, value):
        warnings.warn("SCF_energies is deprecated, use energies.scf instead", DeprecationWarning)
        self.energies.scf = value
        
    @property
    def molecular_orbitals(self):
        warnings.warn("molecular_orbitals is deprecated, use orbitals instead", DeprecationWarning)
        return self.orbitals
    
    @molecular_orbitals.setter
    def molecular_orbitals(self, value):
        warnings.warn("molecular_orbitals is deprecated, use orbitals instead", DeprecationWarning)
        self.orbitals = value
        
    @property
    def spin_orbit_coupling(self):
        warnings.warn("spin_orbit_coupling is deprecated, use soc instead", DeprecationWarning)
        return self.soc
    
    @spin_orbit_coupling.setter
    def spin_orbit_coupling(self, value):
        warnings.warn("spin_orbit_coupling is deprecated, use soc instead", DeprecationWarning)
        self.soc = value
        
    @property
    def dipole_moment(self):
        warnings.warn("dipole_moment is deprecated, use pdm instead", DeprecationWarning)
        return self.pdm
    
    @dipole_moment.setter
    def dipole_moment(self, value):
        warnings.warn("dipole_moment is deprecated, use pdm instead", DeprecationWarning)
        self.pdm = value
        
    @property
    def vertical_emission(self):
        warnings.warn("vertical_emission is deprecated, use emission.vertical instead", DeprecationWarning)
        return self.emission.vertical
    
    @vertical_emission.setter
    def vertical_emission(self, value):
        warnings.warn("vertical_emission is deprecated, use emission.vertical instead", DeprecationWarning)
        self.emission.vertical = value
    
    @property
    def adiabatic_emission(self):
        warnings.warn("adiabatic_emission is deprecated, use emission.adiabatic instead", DeprecationWarning)
        return self.emission.adiabatic
    
    @adiabatic_emission.setter
    def adiabatic_emission(self, value):
        warnings.warn("adiabatic_emission is deprecated, use emission.adiabatic instead", DeprecationWarning)
        self.emission.adiabatic = value
        
    @property
    def metadatas(self):
        """
        Property providing access to the list of metadatas of the calculations that were merged together.
        """
        return [result.metadata for result in self.results]
    
    @property
    def level_of_theory(self):
        """
        A short-hand summary of the methods and basis sets used.
        """
        methods = []
        basis_sets = []
        for metadata in self.metadatas:
            if len(metadata.converted_methods) > 0:
                method = metadata.converted_methods[-1]
#             for method in metadata.converted_methods:
                if method not in methods:
                    methods.append(method)
                
            if metadata.basis_set is not None and metadata.basis_set not in basis_sets:
                basis_sets.append(metadata.basis_set)
                        
        return "/".join(itertools.chain(methods, basis_sets))
    
    @property
    def title(self):
        """
        A string Title describing this result.
        """
        title = ", ".join(self.metadata.calculations)
        if "Excited States" in self.metadata.calculations:
            # Add multiplicity based on ES.
            mult = self.excited_states.group()
            mult_strings = [digichem.result.excited_state.Energy_state.multiplicity_number_to_string(multiplicity).capitalize() for multiplicity in mult]
            if len(mult_strings) <= 2:
                # Add the strings.
                title += " ({})".format(", ".join(mult_strings))
            else:
                # Add something non-specific.
                title += " (Various Multiplicities)"
        else:
            title += " ({})".format(digichem.result.excited_state.Energy_state.multiplicity_number_to_string(self.metadata.multiplicity).capitalize())
        return title
    
    @property
    def energy(self):
        """
        The total energy of this calculation.
        
        This convenience property is the energy at the highest level of theory available (CC > MP > SCF).
        
        :raises Result_unavailable_error: If no total energy is available.
        """
        warnings.warn("energy is deprecated, use energies.final instead", DeprecationWarning)
        return self.energies.final
    
        
    @property
    def electric_transition_dipole_moment(self):
        """
        The S1 electric dipole moment, commonly referred to as THE electric transition dipole moment (although this name is ambiguous).
        
        None is returned if the S1 dipole moment is not available.
        """
        tdm = self.transition_dipole_moment
        if tdm is not None:
            return tdm.electric
        
    @property
    def magnetic_transition_dipole_moment(self):
        """
        The S1 magnetic dipole moment, commonly referred to as THE magnetic transition dipole moment (although this name is ambiguous).
        
        None is returned if the S1 dipole moment is not available.
        """
        tdm = self.transition_dipole_moment
        if tdm is not None:
            return tdm.magnetic

    @property
    def transition_dipole_moment(self):
        """
        The S1 dipole moment (both electric and magnetic), commonly referred to as THE transition dipole moment (although this name is ambiguous).
        
        None is returned if the S1 dipole moment is not available.
        """
        try:
            S1 = self.excited_states.get_state("S(1)")
            return S1.transition_dipole_moment
        except Result_unavailable_error:
            # No S1 available.
            return None
        
    def _dump_(self, digichem_options, all):
        "Dump the data contained in this result set, serialising it to a hierarchy of dicts that can be saved in various formats."
        # Start with our DB ID if we have one.
        dump_dic = {}
        if self._id is not None:
            dump_dic['_id'] = self._id
            
        dump_dic.update({
            "metadata": self.metadata.dump(digichem_options, all),
            "ground_state": self.ground_state.dump(digichem_options, all) if self.ground_state is not None else None,
            "energies": self.energies.dump(digichem_options, all),
            "atoms": self.atoms.dump(digichem_options, all),
            "raw_atoms": self.raw_atoms.dump(digichem_options, all),
            "orbitals": self.orbitals.dump(digichem_options, all),
            "beta_orbitals": self.beta_orbitals.dump(digichem_options, all),
            "pdm": self.pdm.dump(digichem_options, all) if self.pdm is not None else None,
            "excited_states": self.excited_states.dump(digichem_options, all),
            "soc": self.soc.dump(digichem_options, all),
            "vibrations": self.vibrations.dump(digichem_options, all),
            "nmr": self.nmr.dump(digichem_options, all),
            "emission": self.emission.dump(digichem_options, all)
        })
        
        return dump_dic


from digichem.exception import Digichem_exception


class Gaussian_input_parser():
    """
    Class for parsing Gaussian input files.
    """
    
    # Each 'section' in Gaussian is separated by a double newline; this seems to be the only consistent delimiter.
    SECTION_SEPARATOR = "\n\n"
    
    def __init__(self, file_str = None):
        """
        Constructor for Gaussian input files.
        
        :param file_str: String containing a Gaussian input file.
        """
        # Link 0 commands that appear at the top of the input file, we currently do no parsing on these (because we ignore them anyway)
        # Link 0 commands are stored as they appear (as a string, possibly containing newlines).
        self.link_0 = None
        
        # The route section, describes the calculation to be performed.
        ## Most of the route is ignored as we set it ourselves as part of the submission process, but some options (geom=connectivity, genECP etc) control the format of the rest of the file, so we do parse these.
        ## This is a dictionary of options; where 'key: value' translates to 'key=value' in the gaussian file. If value is None, then the translation is to 'key'.
        # The route section is currently not parsed (it is just a string).        
        self.route = None
        
        # The title of the input file.
        self.title = None
        
        # The multiplicity and charge of the molecule, appears at the top of the geometry section as 'charge, mult' or 'charge mult'.
        self.multiplicity = None
        self.charge = None
        
        # The geometry (atoms and charge etc) section as a string (almost certainly containing newlines).
        self.geometry = None
        
        # These are additional sections that can optionally appear in some input files (connectivity, basis set etc).
        self.additional_sections = []
        # If we've been given a file, load it now.
        if file_str is not None:
            self.load(file_str)
        
    @property
    def title(self):
        """
        Get the title section of this input file.
        
        None and empty string values are translated to a single whitespace character (because otherwise they will be interpreted wrong).
        """
        if self._title is not None and len(self._title.trim()) != 0:
            return self._title
        
        else:
            return "(title)"
    
    @title.setter
    def title(self, value):
        """
        Set the title section of this input file.
        """
        self._title = value
    
    def load(self, file_str):
        """
        Load a Gaussian input file.
        
        :param file_str: String containing a Gaussian input file.
        
        """
        # First, split on our delimeter.
        sections = file_str.split("\n\n")
        
        link_0_lines = []
        route_lines = []
        # Split the first section into link 0 and route (link 0 starts with %, the first line not to start without % is route).
        if len(sections) > 0:
            in_route = False
            for line in sections[0].split("\n"):
                # Check first char.
                if not in_route and line[:1] != "%":
                    in_route = True
                    
                # Add to one of our two lists.
                if not in_route:
                    link_0_lines.append(line)
                else:
                    route_lines.append(line)
        
        # Now set.
        self.link_0 = "\n".join(link_0_lines) if len(link_0_lines) > 0 else None
        self.route = "\n".join(route_lines) if len(route_lines) > 0 else None
        
        # Set title.
        self.title = sections[1] if len(sections) > 1 else None
        
        # Geometry (the first line contains charge and mult).
        try:
            geometry_section = sections[2].split("\n", 1)
        except IndexError:
            # No geometry available.
            raise Digichem_exception("Failed to read Gaussian geometry data from input file; is the file formatted correctly?") 
        
        # We'll first try to split on comma (,) for charge,mult.
        charge_mult = geometry_section[0].split(",", 1)
        # If we didn't get what we want, try on whitespace.
        if len(charge_mult) != 2:
            charge_mult = geometry_section[0].split(" ", 1)
        
        # Now try and set.
        try:
            self.charge = int(charge_mult[0])
            self.multiplicity = int(charge_mult[1])
        except Exception:
            raise Digichem_exception("Unable to determine charge and multiplicity from '{}'".format(charge_mult))
        
        # The rest of the geometry section contains atoms.
        if len(geometry_section) != 2 or geometry_section[1] == "":
            raise Digichem_exception("Gaussian input file does not appear to contain any atoms")
        
        self.geometry = geometry_section[1]
        
        # Run a sanity check on the geometry format.
        for geometry_line in self.geometry.split("\n"):
            num_columns = len(geometry_line.split())
            if num_columns > 4:
                # We have too many columns?
                raise Digichem_exception("Gaussian input file has too many columns ({}) in its geometry section".format(num_columns))
                
        
        # And anything else.
        self.additional_sections = sections[3:]
        
    @property
    def xyz(self):
        """
        Get the geometry of this gaussian input file in XYZ format.
        """
        return "{}\n\n{}".format(len(self.geometry.split("\n")), self.geometry)




"""Test for memory representations."""

import pytest

from digichem.memory import Memory


@pytest.mark.parametrize("value, valid", [
        ["1", True],
        [1, True],
        [1.0, True],
        ["1 B", True],
        ["1 KB", True],
        ["1 MB", True],
        ["1 GB", True],
        ["1 TB", True],
        ["foo bar", False],
     ])
def test_validation(value, valid):
    try:
        mem = Memory(value)

        str(mem)
    
    except Exception:
        if valid:
            raise
        
        else:
            return
    
    if not valid:
        raise


"""
Classes for parsing calculation result data.

Most of the heavy lifting is done by cclib, we just extract additional data not currently handed by cclib.
"""

# We're not bothered about these warnings.
import warnings
warnings.filterwarnings("ignore", "Module .* is under testing", category = UserWarning)
warnings.filterwarnings("ignore", "Module .* is not fully tested", category = UserWarning)

# These alignment classes are needed to parse correctly.
from digichem.result.alignment.AAA import Adjusted_average_angle
from digichem.result.alignment.AA import Average_angle
from digichem.result.alignment.FAP import Furthest_atom_pair
from digichem.result.alignment import Minimal

from digichem.parse.util import from_log_files, parse_calculation, parse_and_merge_calculations, parse_multiple_calculations, parse_and_merge_multiple_calculations, open_for_parsing


# Import top-level object for easier access.
from .base import Digichem_exception
from .base import Result_unavailable_error
from .base import File_maker_exception
from .base import Unknown_file_type_exception
from .base import Submission_error
from .base import Format_error


"""Tests for checking the values of parsed results."""

import pytest
from pathlib import Path
import scipy.constants

from digichem.parse import parse_calculation
from digichem.parse.util import parse_and_merge_calculations
from digichem.test.util import data_directory, digichem_options
from digichem.test.util import check_float_list, check_dipole, check_orbitals


@pytest.fixture(scope="module")
def gaussian_SP_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Naphthalene/Gaussian 16 Single Point (Singlet) PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def gaussian_opt_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Naphthalene/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Toluene 6-31G(d,p)"), options = digichem_options)

@pytest.fixture(scope="module")
def gaussian_radical_anion_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Benzene Anion/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Gas Phase 6-31G(d,p).tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def gaussian_ES_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Naphthalene/Gaussian 16 Excited States TDA 10 Singlets 10 Triplets PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def gaussian_emission_result(digichem_options):
    return parse_and_merge_calculations(
        Path(data_directory(), "Naphthalene/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Toluene 6-31G(d,p)"),
        Path(data_directory(), "Naphthalene/Gaussian 16 Excited States TDA Optimised S(1) PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"),
        options = digichem_options
    )

@pytest.fixture(scope="module")
def gaussian_opt_ES_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Naphthalene/Gaussian 16 Excited States TDA Optimised S(1) PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def gaussian_PDM_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def gaussian_TDM_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Gaussian 16 Excited States TDA 10 Singlets 10 Triplets PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def gaussian_PDM_ES_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Gaussian 16 Excited States TDA Optimised S(1) PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"), options = digichem_options)

#############
# Turbomole #
#############

@pytest.fixture(scope="module")
def turbomole_sp_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Turbomole Single Point DFT PBE0 (GD3BJ) Toluene Pople Basis Sets STO-3G.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def turbomole_opt_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Naphthalene/Turbomole Optimisation Frequency PBE0 (GD3BJ) 6-31G**.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def turbomole_ES_singlets_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Naphthalene/Turbomole Excited States TDA 10 Singlets PBE0 (GD3BJ) 6-31G**.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def turbomole_ES_triplets_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Naphthalene/Turbomole Excited States TDA 10 Triplets PBE0 (GD3BJ) 6-31G**.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def turbomole_ES_result(digichem_options):
    return parse_and_merge_calculations(
        Path(data_directory(), "Naphthalene/Turbomole Excited States TDA 10 Singlets PBE0 (GD3BJ) 6-31G**.tar.gz"),
        Path(data_directory(), "Naphthalene/Turbomole Excited States TDA 10 Triplets PBE0 (GD3BJ) 6-31G**.tar.gz"),
        options = digichem_options
    )

@pytest.fixture(scope="module")
def turbomole_PDM_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Turbomole Optimisation Frequency PBE0 (GD3BJ) 6-31G**.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def turbomole_TDM_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Turbomole Excited States TDA 10 Singlets PBE0 (GD3BJ) 6-31G**.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def turbomole_radical_anion_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Benzene Anion/Turbomole Optimisation Frequency PBE0 (GD3BJ) 6-31G**.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def turbomole_ADC2_opt_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Naphthalene/Turbomole Optimisation ADC(2) cc-pVDZ.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def turbomole_ADC2_singlets_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Naphthalene/Turbomole Excited States ADC(2) S(1) and S(2) cc-pVDZ.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def turbomole_ADC2_triplets_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Naphthalene/Turbomole Excited States ADC(2) T(1) and T(2) cc-pVDZ.tar.gz"), options = digichem_options)

########
# ORCA #
########

@pytest.fixture(scope="module")
def orca_SP_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Orca Single Point PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def orca_solvent_SP_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Orca Single Point PBE0 (GD3BJ) Toluene Pople Basis Sets STO-3G.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def orca_grad_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Orca Gradient PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def orca_opt_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Orca Optimisation PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def orca_opt_freq_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Orca Optimisation Frequencies PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def orca_freq_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Orca Frequencies PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def orca_nmr_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Orca NMR PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz"), options = digichem_options)

@pytest.fixture(scope="module")
def orca_ES_result(digichem_options):
    return parse_calculation(Path(data_directory(), "Pyridine/Orca Excited States TDA 10 Singlets 10 Triplets PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz"), options = digichem_options)


@pytest.mark.parametrize("result_set, num, final", [
        (pytest.lazy_fixture("gaussian_SP_result"), 1, -10488.990333747),
        (pytest.lazy_fixture("gaussian_opt_result"), 5, -10488.990333747),
        (pytest.lazy_fixture("gaussian_ES_result"), 1, -10488.990333747),
        (pytest.lazy_fixture("gaussian_opt_ES_result"), 4, -10488.883505795338),
        (pytest.lazy_fixture("turbomole_ADC2_opt_result"), 7, -10432.311381999756),
        (pytest.lazy_fixture("turbomole_ADC2_singlets_result"), 1, -10432.311386936175),
        (pytest.lazy_fixture("turbomole_ADC2_triplets_result"), 1, -10432.311386936175)
    ])
def test_SCF_energy(result_set, num, final):
    """Test the parsed energy is correct"""
    assert result_set.energies.scf.final == pytest.approx(final, abs=1e-5)
    
    # Check length, which will be 1 for SP, and >1 for the opts.
    assert len(result_set.energies.scf) == num


@pytest.mark.parametrize("result_set, num, final", [
        (pytest.lazy_fixture("turbomole_ADC2_opt_result"), 7, -10467.158186939303),
        (pytest.lazy_fixture("turbomole_ADC2_singlets_result"), 1, -10467.158181211305),
        (pytest.lazy_fixture("turbomole_ADC2_triplets_result"), 1, -10467.158181211305)
    ])
def test_MP_energy(result_set, num, final):
    """Test the parsed energy is correct"""
    assert result_set.energies.mp.final == pytest.approx(final, abs=1e-5)
    
    # Check length, which will be 1 for SP, and >1 for the opts.
    assert len(result_set.energies.scf) == num
    assert len(result_set.energies.mp) == num


@pytest.mark.parametrize("result_set, charge, mult, energy", [
        (pytest.lazy_fixture("gaussian_opt_result"), 0, 1, -10488.990333747),
        (pytest.lazy_fixture("turbomole_ADC2_opt_result"), 0, 1, -10467.158186939303),
        # Radical anions.
        (pytest.lazy_fixture("gaussian_radical_anion_result"), -1, 2, -6310.5380531853125),
        (pytest.lazy_fixture("turbomole_radical_anion_result"), -1, 2, -6310.4572047072043),
    ])
def test_ground_state(result_set, charge, mult, energy):
    """Test the ground state properties"""
    
    assert result_set.ground_state.charge == charge
    assert result_set.ground_state.multiplicity == mult
    # Ground states have zero excited state energy by definition.
    assert result_set.ground_state.energy == 0.0
    # This is the total energy.
    assert result_set.ground_state.absolute_energy == pytest.approx(energy, abs=1e-5)


@pytest.mark.parametrize("result_set", [
        pytest.lazy_fixture("gaussian_SP_result"),
        pytest.lazy_fixture("gaussian_opt_result"),
        pytest.lazy_fixture("gaussian_ES_result"),
        pytest.lazy_fixture("gaussian_opt_ES_result"),
        pytest.lazy_fixture("turbomole_ADC2_opt_result"),
        pytest.lazy_fixture("turbomole_ADC2_singlets_result"),
        pytest.lazy_fixture("turbomole_ADC2_triplets_result"),
    ])
def test_atoms(result_set):
    """Test the parsed (unaligned) atoms are correct."""
    # The molecule is Naphthalene; C10H8
    # Check num atoms.
    assert len(result_set.raw_atoms) == 18
    
    # Check atom types.
    assert result_set.raw_atoms.element_dict['C'] == 10
    assert result_set.raw_atoms.element_dict['H'] == 8
    
    # Check mass.
    assert result_set.raw_atoms.molar_mass == pytest.approx(128.174)
    
    # We don't check positions here because these can vary from calc to calc due to reorientation...


@pytest.mark.parametrize("result_set, length, width, height", [
        (pytest.lazy_fixture("gaussian_SP_result"), 6.739394, 4.972394, 0.000004),
        (pytest.lazy_fixture("gaussian_opt_result"), 6.739394, 4.972394, 0.000004),
        (pytest.lazy_fixture("gaussian_ES_result"), 6.739394, 4.972394, 0.000004),
        (pytest.lazy_fixture("turbomole_ADC2_opt_result"), 6.8000171599, 5.0161488999, 0.0000475622),
        (pytest.lazy_fixture("turbomole_ADC2_singlets_result"), 6.8000171599, 5.0161488999, 0.0000475622),
        (pytest.lazy_fixture("turbomole_ADC2_triplets_result"), 6.8000171599, 5.0161488999, 0.0000475622),
    ])
def test_alignment(result_set, length, width, height):
    """Test the aligned atom positions are correct."""
    # Rounding errors plague parsing and conversion of atomic coordinates.
    # Much of this error comes from the accuracy printed by the CC programs themselves,
    # so there is not much we can do about it.
    tolerance = 1e-5
    
    # Dimensions in angstrom.
    try:
        linear_ratio = 1-(width/length)
    except (FloatingPointError, ZeroDivisionError):
        linear_ratio = 1.0
        
    try:
        planar_ratio = 1-(height/width)
    except (FloatingPointError, ZeroDivisionError):
        planar_ratio = 1.0
    
    # Check dimensions.
    assert result_set.atoms.X_length == pytest.approx(length, abs=tolerance)
    assert result_set.atoms.Y_length == pytest.approx(width, abs=tolerance)
    assert result_set.atoms.Z_length == pytest.approx(height, abs=tolerance)
    
    # Check params.
    assert result_set.atoms.get_linear_ratio() == pytest.approx(linear_ratio, abs=tolerance)
    assert result_set.atoms.get_planar_ratio() == pytest.approx(planar_ratio, abs=tolerance)


@pytest.mark.parametrize("result_set, num_occ, num_unocc, homo, lumo", [
        (pytest.lazy_fixture("gaussian_opt_result"), 34, 156, -6.13072481, -0.92437071),
        (pytest.lazy_fixture("turbomole_ADC2_opt_result"), 34, 146, -7.78345591, 2.37048311)
    ])
def test_orbitals(result_set, num_occ, num_unocc, homo, lumo):
    """Test the energies of the parsed MOs"""
    check_orbitals(result_set.orbitals, num_occ, num_unocc, homo, lumo)
    
    # Check spin labels
    assert result_set.orbitals.spin_type == "none"


@pytest.mark.parametrize("result_set, num_occ, num_unocc, homo, lumo", [
        (pytest.lazy_fixture("gaussian_radical_anion_result"), (22, 21), (98, 99), (3.832451319, -0.0157826027), (6.268142298, 6.34270149)),
        (pytest.lazy_fixture("turbomole_radical_anion_result"), (22, 21), (92, 93), (3.825336056180260069, 0.0034865153882969327284), (6.275672672175910627, 6.387468734183322283))
    ])
def test_unrestricted_orbitals(result_set, num_occ, num_unocc, homo, lumo):
    """Test the energies of unrestricted orbitals"""
    # Check overall length.
    assert len(result_set.orbitals) == len(result_set.beta_orbitals)
    
    # Check each set of orbitals.
    check_orbitals(result_set.orbitals, num_occ[0], num_unocc[0], homo[0], lumo[0])
    check_orbitals(result_set.beta_orbitals, num_occ[1], num_unocc[1], homo[1], lumo[1])
    
    # Check spin labels
    assert result_set.orbitals.spin_type == "alpha"
    assert result_set.beta_orbitals.spin_type == "beta"


@pytest.mark.parametrize("result_set, num, index, frequency, intensity", [
        (pytest.lazy_fixture("gaussian_opt_result"), 48, 13, 806.2853, 121.0015),
        (pytest.lazy_fixture("gaussian_SP_result"), 0, None, None, None),
        (pytest.lazy_fixture("turbomole_opt_result"), 48, 13, 805.28, 103.77),
    ])
def test_frequencies(result_set, num, index, frequency, intensity):
    """Test the vibrational frequencies are correct"""
    
    # Check lengths.
    assert len(result_set.vibrations) == num
    assert len(result_set.vibrations.negative_frequencies) == 0

    if index != None:
        # Check one of the energies.
        assert result_set.vibrations[index].frequency == pytest.approx(frequency)
        assert result_set.vibrations[index].intensity == pytest.approx(intensity)
        
        # Check the frequencies are ordered.
        indices = [freq.level for freq in result_set.vibrations]
        assert indices == list(range(1, num+1))
        
    else:
        with pytest.raises(IndexError):
            result_set.vibrations[-1].frequency


@pytest.mark.parametrize("result_set, coords, axis_angle, plane_angle", [
        (pytest.lazy_fixture("gaussian_PDM_result"), (0.0, 2.5103, 0.0), 90.0, 0.0),
        (pytest.lazy_fixture("gaussian_PDM_ES_result"), (0.0001, -0.6147, 0.0001), 90.0, 0.0),
        (pytest.lazy_fixture("turbomole_PDM_result"), (-0.000743401, -2.20405438, -0.000494493), 90.0, 0.0),
    ])
def test_pdm(result_set, coords, axis_angle, plane_angle):
    """Test the permanent dipole moment"""
     
    check_dipole(result_set.pdm, coords)
    
    # Check angles
    assert float(result_set.pdm.X_axis_angle) == pytest.approx(axis_angle, abs=1e-1)
    assert float(result_set.pdm.XY_plane_angle) == pytest.approx(plane_angle, abs=1e-1)


@pytest.mark.parametrize("result_set, number, TDM", [
        (pytest.lazy_fixture("gaussian_TDM_result"), 20, (0.0, -0.0003, 0.5879)),
        (pytest.lazy_fixture("turbomole_TDM_result"), 10, (-0.000628, -0.000132, 0.558096))
    ])
def test_tedm(result_set, number, TDM):
    """Test transition dipole moments"""
    # Check number of dipoles.
    assert len([excited_state.transition_dipole_moment for excited_state in result_set.excited_states if excited_state.transition_dipole_moment.electric is not None]) == number
    
    # Check the S1 moments are correct.
    S1_TDM = result_set.excited_states.get_state("S(1)").transition_dipole_moment
    
    # Check coords and magnitude.
    check_dipole(S1_TDM.electric, TDM)


@pytest.mark.parametrize("result_set, number, TDM", [
        (pytest.lazy_fixture("gaussian_TDM_result"), 20, (0.6909, 0.0, 0.0)),
        (pytest.lazy_fixture("turbomole_TDM_result"), 10, (0.673218, -0.000274, -0.000274))
    ])
def test_tmdm(result_set, number, TDM):
    """Test transition dipole moments"""
    
    # Check number of dipoles.
    assert len([excited_state.transition_dipole_moment for excited_state in result_set.excited_states if excited_state.transition_dipole_moment.magnetic is not None]) == number
    
    # Check the S1 moments are correct.
    S1_TDM = result_set.excited_states.get_state("S(1)").transition_dipole_moment
    
    # Check coords and magnitude.
    check_dipole(S1_TDM.magnetic, TDM, abs = 1e-3)


@pytest.mark.parametrize("result_set, state, g_lum", [
        (pytest.lazy_fixture("gaussian_TDM_result"), "S(5)", -0.995)
    ])
def test_g_lum(result_set, state, g_lum):
    """Test calculation of G(lum) values for circularly polarized light."""
    
    TDM = result_set.excited_states.get_state(state).transition_dipole_moment
    
    TEDM = TDM.electric.gaussian_cgs_vector
    TMDM = TDM.magnetic.gaussian_cgs_vector
    
    # Test values in Gaussian-CGS units.))
    check_float_list(TEDM, tuple(254.35e-20 * i for i in TEDM))
    check_float_list(TMDM, tuple(-0.97401e-20 * i for i in TMDM))
    
    # Check angle.
    angle = (TEDM[0] * TMDM[0] + TEDM[1] * TMDM[1] + TEDM[2] * TMDM[2]) / ( ((TEDM[0]**2 + TEDM[1] **2 + TEDM[2] **2)**0.5) * ((TMDM[0]**2 + TMDM[1] **2 + TMDM[2] **2)**0.5) )
    assert TDM.cos_angle() == pytest.approx(angle)
    
    # Check g_lum
    assert TDM.g_value == pytest.approx(g_lum, abs=1e-3)


@pytest.mark.parametrize("result_set, num_singlets, num_triplets, dest, state_label, state_index, state_symmetry, state_energy, state_f", [
        (pytest.lazy_fixture("gaussian_ES_result"), 10, 10, 1.6231, "S(2)", 7, "Singlet-B1u", 4.7387, 0.1168),
        (pytest.lazy_fixture("turbomole_ES_result"), 10, 10, 1.6298, "S(2)", 7, "Singlet-A", 4.790868037954573, 0.0805)
    ])
def test_excited_states(result_set, num_singlets, num_triplets, dest, state_label, state_index, state_symmetry, state_energy, state_f):
    """Test excited states"""
    
    # Check number of states.
    assert result_set.excited_states.num_singlets == num_singlets
    assert result_set.excited_states.num_triplets == num_triplets
    
    # Check singlet-triplet splitting energy (dE(ST)).
    assert result_set.excited_states.singlet_triplet_energy == pytest.approx(dest, abs=1e-4)
    
    # Now check a specific state.
    state = result_set.excited_states.get_state(state_label)
    assert state.level == state_index
    assert state.symmetry == state_symmetry
    assert state.energy == pytest.approx(state_energy, abs=1e-4)
    assert state.oscillator_strength == pytest.approx(state_f, abs=1e-4)
    
    # Check wavelength.
    # e in J
    joule_energy = state_energy * scipy.constants.physical_constants['electron volt'][0]
    # wavelength in m
    meter_wavelength = (scipy.constants.Planck * scipy.constants.c) / joule_energy
    # wavelength in nm
    wavelength = meter_wavelength *1000000000
    
    assert state.wavelength == pytest.approx(wavelength, abs=1e-4)


@pytest.mark.parametrize("result_set, state_index, transition_index, start_orbital, end_orbital, coefficient", [
        (pytest.lazy_fixture("gaussian_ES_result"), 1, 1, 34, 35, 0.957238734),
        (pytest.lazy_fixture("gaussian_ES_result"), 1, 2, 32, 37, 0.186520627),
        (pytest.lazy_fixture("gaussian_ES_result"), 1, 3, 33, 36, 0.175928167),
        (pytest.lazy_fixture("turbomole_ES_result"), 2, 1, 33, 35, 0.772657751),
        (pytest.lazy_fixture("turbomole_ES_result"), 2, 2, 34, 36, 0.618869938),
    ])
def test_excited_state_transitions(result_set, state_index, transition_index, start_orbital, end_orbital, coefficient):
    """Test excited state transitions."""
    
    state = result_set.excited_states[state_index -1]
    transition = state.transitions[transition_index -1]
    
    assert transition.level == transition_index
    assert transition.starting_mo.level == start_orbital
    assert transition.ending_mo.level == end_orbital
    assert transition.coefficient == pytest.approx(coefficient)
    assert transition.probability == pytest.approx(coefficient **2)


@pytest.mark.parametrize("result_set, transition_type, multiplicity, emission_type, energy, excited_energy, oscillator_strength, rate", [
        (pytest.lazy_fixture("gaussian_emission_result"), "adiabatic", 1, "fluorescence", 4.542828007, -10484.452882954, 0.0, 312658.74889),
        (pytest.lazy_fixture("gaussian_emission_result"), "vertical", 1, "fluorescence", 4.4360, -10484.452882954, 0.0, 291116.171714),
    ])
def test_emission(result_set, transition_type, multiplicity, emission_type, energy, excited_energy, oscillator_strength, rate):
    # TODO: Testing of fluorescence rate should be moved to another module (where other calculated properties can be tested).
    if transition_type == "adiabatic":
        emission = result_set.emission.adiabatic[multiplicity]
        
    else:
        emission = result_set.emission.vertical[multiplicity]
        
    # Test some general properties.
    assert emission.transition_type == transition_type
    assert emission.emission_type == emission_type
    
    # Test numerical properties.
    assert emission.energy == pytest.approx(energy)
    assert emission.excited_energy == pytest.approx(excited_energy)
    assert emission.oscillator_strength == pytest.approx(oscillator_strength)
    assert emission.emission_rate == pytest.approx(rate)


def test_h_nmr_standard(orca_nmr_result):
    reference = 31.68766666666667
    # Check the reference standard is available.
    assert all([nmr_result.shielding.reference == pytest.approx(reference, abs = 1e-4) for nmr_result in orca_nmr_result.nmr if nmr_result.atom.element.symbol == "H"])

def test_c_nmr_standard(orca_nmr_result):
    reference = 197.90316666666664
    # Check the reference standard is available.
    assert all([nmr_result.shielding.reference == pytest.approx(reference, abs = 1e-4) for nmr_result in orca_nmr_result.nmr if nmr_result.atom.element.symbol == "C"])

def test_nmr_h_isotope_options(orca_nmr_result):
    # Check we have the right options set for H.
    options = orca_nmr_result.nmr.spectrometer.isotope_options(1, 1)
    assert options['frequency'] == 400
    assert options['fwhm'] == 0.005
    assert options['gaussian_resolution'] == 0.0005
    assert options['coupling_filter'] == 0.001
    assert options['pre_merge'] == 0.0005

def test_nmr_c_isotope_options(orca_nmr_result):
    # Check we have the right options set for H.
    options = orca_nmr_result.nmr.spectrometer.isotope_options(6, 13)
    assert options['frequency'] == 100.6

@pytest.mark.parametrize("result_set", [
        pytest.lazy_fixture("gaussian_SP_result"),
        pytest.lazy_fixture("gaussian_emission_result"),
        pytest.lazy_fixture("turbomole_sp_result"),
        pytest.lazy_fixture("orca_solvent_SP_result"),
    ])
def test_metadata_solvent(result_set):
    """Can we parse solution metadata correctly?"""
    assert result_set.metadata.solvent.name == "Toluene"
    # A bit of a rudimentary test, but there's already more advanced testing in cclib
    assert result_set.metadata.solvent.model is not None
    assert result_set.metadata.solvent.params['epsilon'] == pytest.approx(2.374, abs = 1e-3)



import pytest
from pathlib import Path
import shutil

from digichem.file.fchk import Chk_to_fchk
from digichem.file.cube import Fchk_to_cube
from digichem.parse.util import open_for_parsing, parse_calculation
from digichem.exception import File_maker_exception

from digichem.test.util import digichem_options, data_directory
from digichem.test.test_result import gaussian_ES_result, turbomole_ES_result, orca_ES_result, \
    gaussian_opt_result, turbomole_opt_result, orca_opt_result, orca_opt_freq_result, \
    orca_nmr_result

@pytest.mark.skipif(not shutil.which("cubegen"),
                    reason="No cubegen available")
@pytest.mark.parametrize("result_path", [
    Path(data_directory(), "Pyridine/Gaussian 16 Excited States TDA Optimised S(1) PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"),
], ids=["Pyridine"])
@pytest.mark.parametrize("sanity", [False, True], ids=["Normal", "Sanitize"])
def test_cube_from_fchk(result_path, sanity, tmp_path, digichem_options):
    """Can we make a cube file from an fchk file?"""
    with open_for_parsing(result_path) as open_files:
        result_set = parse_calculation(*open_files, options = digichem_options)
        maker = Fchk_to_cube.from_options(
            tmp_path / "tmp.cube",
            fchk_file = result_set.metadata.auxiliary_files['fchk_file'],
            options = digichem_options,
        )
        maker.sanitize = sanity

        assert maker.get_file().exists()

        # Check the cube is formatted correctly.
        with open(maker.get_file()) as cube:
            line = next(cube)
            line = next(cube)
            line = next(cube)
            split_line = line.split()
            if sanity:
                assert len(split_line) == 4
            
            else:
                assert len(split_line) == 5


def test_no_cubegen_error(digichem_options, tmp_path):
    """Do we throw the right exception if cubegen doesn't exist?"""
    maker = Fchk_to_cube.from_options(
        tmp_path / "tmp.cube",
        fchk_file = "foobar",
        options = digichem_options,
    )
    maker.cubegen_executable = "NOTREAL"

    with pytest.raises(File_maker_exception):
        maker.get_file()

def test_no_fchk_error(digichem_options, tmp_path):
    """Do we throw the right exception if we don't give an fchk file?"""
    maker = Fchk_to_cube.from_options(
        tmp_path / "tmp.cube",
        fchk_file = "foobar",
        options = digichem_options,
    )

    with pytest.raises(File_maker_exception):
        maker.get_file()

@pytest.mark.skipif(not shutil.which("formchk"),
                    reason="No formchk available")
def test_chk_to_fchk(digichem_options, tmp_path):
    """Can we convert a chk to an fchk file?"""
    maker = Chk_to_fchk.from_options(
        tmp_path / "tmp.fchk",
        chk_file = Path(data_directory(), "Chk/Pyridine.opt.chk"),
        options = digichem_options
    )
    assert maker.get_file().exists()


def test_no_formchk_error(digichem_options, tmp_path):
    """Do we throw the right exception if formchk doesn't exist?"""
    maker = Chk_to_fchk.from_options(
        tmp_path / "tmp.fchk",
        chk_file = Path(data_directory(), "Input/Benzene.chk"),
        options = digichem_options
    )
    maker.formchk_executable = "NOTREAL"

    with pytest.raises(File_maker_exception):
        maker.get_file()


def test_no_chk_error(digichem_options, tmp_path):
    """Do we throw the right exception if we don't give a chk file?"""
    maker = Chk_to_fchk.from_options(
        tmp_path / "tmp.fchk",
        chk_file = "foobar",
        options = digichem_options
    )

    with pytest.raises(File_maker_exception):
        maker.get_file()


"""Tests for translation classes."""
import pytest

from digichem.translate import Solvent, Basis_set, Multiplicity, Functional,\
    Cube_grid_points


def test_solvent():
    """
    Test conversions for DCM.
    """
    # Get some different ways to represent DCM.
    solvents = (Solvent("dcm"), Solvent("DCM"), Solvent("dichloroMethane"), Solvent("Dichloro Methane"))
    
    # Check all have equivalent string representations.
    assert all([str(solvents[0]) == str(solvent) for solvent in solvents])
    
    # And the correct gaussian name.
    assert all(["DiChloroMethane" == solvent.to_gaussian() for solvent in solvents])
    
    # And epsilon
    assert all([pytest.approx(8.93) == solvent.translate("epsilon") for solvent in solvents])
    

def test_631g_str_str():
    """
    Test conversions for basis sets.
    """
    # Test both types of 6-31G**.
    basis_sets = (Basis_set("6-31g(d,p)"), Basis_set("6-31G**"))
    
    # Check both are equivalent.
    assert str(basis_sets[0]) == str(basis_sets[1])
    
    # Check Gaussian version.
    assert all(["6-31G**" == basis_set.to_gaussian() for basis_set in basis_sets])
    
    # Check Turbomole version.
    assert all(["6-31G**" == basis_set.to_turbomole() for basis_set in basis_sets])
    
def test_def2_svp_p():
    """
    Test weird gaussian name for def2-SV(P).
    """
    assert Basis_set("def2-SV(P)").to_gaussian() == "def2SVPP"

@pytest.mark.parametrize("basis_set_name", ["def2-SVP", "def2-TZV", "def2-TZVP", "def2-QZV", "def2-QZVP"])
def test_karlsruhe(basis_set_name):
    """
    Test weird gaussian names for Karlsruhe basis sets.
    """
    assert "-" not in Basis_set(basis_set_name).to_gaussian()
    
def test_ccpvdz():
    assert Basis_set("cc-PvDz").to_gaussian() == "cc-pVDZ"
    assert str(Basis_set("cc-PvDz")) == "cc-pVDZ"
    

def test_b3lyp():
    """
    Test conversions for DFT functionals.
    """
    funcs = (Functional("b3-LYp"), Functional("B3lyp"))
    
    assert all(["b3-lyp" == func.to_turbomole() for func in funcs])
    
    assert all(["B3LYP" == func.to_gaussian() for func in funcs])


def test_pbe0gaussian():
    # Turbomole functionals are case sensitive. This is particularly annoying because functional
    # names are all lower-case, except for the word 'Gaussian'. Check this.
    assert Functional("PBE0 gaussian").to_turbomole() == "pbe0 Gaussian"


def test_multiplicity():
    """
    Test conversion for multiplicity.
    """
    mults = (Multiplicity(1), Multiplicity("1"), Multiplicity("singlet"), Multiplicity("sInGlEt"), Multiplicity("S"))
    
    # Check conversions.
    assert all(["Singlet" == mult.to_gaussian() for mult in mults])
    
    assert all([1 == mult.to_turbomole() for mult in mults])
    
    assert all(["S" == mult.symbol for mult in mults])
    
    assert all([1 == mult.number for mult in mults])
    
    assert all(["Singlet" == mult.string for mult in mults])
    
def test_grid_points():
    """
    Test different names for cube grid points.
    """
    # Check the special 'Default' value is correct.
    default = Cube_grid_points("default")
    
    assert default.translate("gaussian")    == 0
    assert default.translate("orca")        == 100
    assert default.translate("turbomole")   == "m3"
    
    # Check the named values.
    for name, value in [("Tiny", 25), ("Small", 50), ("Medium", 100), ("Large", 200), ("Huge", 500)]:
        assert Cube_grid_points(name).translate("points") == value
        
    # Check a random value.
    assert Cube_grid_points(123).translate("points") == 123






"""Common testing utilities and convenience functions."""

import pytest
from pathlib import Path

import digichem.config
from digichem.datas import get_resource


def data_directory():
    """
    Get a path to the test data directory.
    """
    return get_resource('test/data')


# Input files.
benzene_cdx = Path(data_directory(), "Input", "Benzene.cdx")
pyridine_cml = Path(data_directory(), "Input", "Pyridine.cml")
cyclopentane_com = Path(data_directory(), "Input", "Cyclopentane.com")
water_xyz = Path(data_directory(), "Input", "Water.xyz")
ethane_xyz = Path(data_directory(), "Input", "Ethane.xyz")
pyridine_si_v2 = Path(data_directory(), "Input/Pyridine.v2.si")
pyridine_si_v1 = Path(data_directory(), "Input/Pyridine.v1.si")


# Result files for testing parsing.
result_files = {
    "gaussian": [Path(data_directory(), datum) for datum in [
        'Naphthalene/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Toluene 6-31G(d,p)',
        'Naphthalene/Gaussian 16 Excited States TDA 10 Singlets 10 Triplets PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz',
        'Naphthalene/Gaussian 16 Excited States TDA Optimised S(1) PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz',
        'Pyridine/Gaussian 16 Excited States TDA Optimised S(1) PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz'
    ]],
    "turbomole": [Path(data_directory(), datum) for datum in [
        'Naphthalene/Turbomole Optimisation ADC(2) cc-pVDZ.tar.gz',
        'Naphthalene/Turbomole Excited States ADC(2) S(1) and S(2) cc-pVDZ.tar.gz',
        'Naphthalene/Turbomole Excited States ADC(2) T(1) and T(2) cc-pVDZ.tar.gz',
    ]],
    "orca": [Path(data_directory(), datum) for datum in [
        'Pyridine/Orca Single Point PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz',
        'Pyridine/Orca Gradient PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz',
        'Pyridine/Orca Optimisation PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz',
        'Pyridine/Orca Optimisation Frequencies PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz',
        'Pyridine/Orca Frequencies PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz',
        'Pyridine/Orca NMR PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz',
        'Pyridine/Orca Excited States TDA 10 Singlets 10 Triplets PBE0 (GD3BJ) Gas Phase Pople Basis Sets STO-3G.tar.gz'
    ]]
}

@pytest.fixture(scope="package")
def digichem_options(tmpdir_factory):
    config = digichem.config.get_config()
    
    # Change the default DB locations to prevent SPAM.
    config['databases'] = [
        "{{name: test1, path: {}}}".format(
            Path(tmpdir_factory.mktemp("silico_database1"), "tmp.db")
        ),
        "{{name: test2, path: {}}}".format(
            Path(tmpdir_factory.mktemp("silico_database2"), "tmp.db")
        ),
        "{{name: test3, auto_fill: False, path: {}}}".format(
            Path(tmpdir_factory.mktemp("silico_database3"), "tmp.db")
        ),
    ]
    config.logging['render_logging'] = True
    # Set blender quality as low as feasible to shorten test time.
    config.render['batoms']['render_samples'] = 1
    config.validate()
    return config


def check_float_list(test_list, expected_list, abs = 1e-4):
    """
    Helper function to compare to lists of floats.
    """
    # Pytest bug #9921 prevents this comparison from working for now...
    #assert test_list == pytest.approx(expected_list, abs=abs)
    for list_index in range(0,len(expected_list)):
        assert test_list[list_index] == pytest.approx(expected_list[list_index], abs=abs)


def check_dipole(dipole_moment, coords, abs = 1e-4):
    """
    Helper function to check a dipole moment (PDM, TEDM or TMDM) matches some expected values.
    """
    check_float_list(dipole_moment.vector_coords, coords, abs)
        
    # Check total
    assert dipole_moment.total == pytest.approx((coords[0] **2 + coords[1] **2 + coords[2] **2) **0.5, abs = abs)


def check_orbitals(orbitals, num_occ, num_unocc, homo, lumo):
    """Helper function for checking orbital energies"""
    # Check numbers.
    assert len(orbitals) == num_occ + num_unocc
    assert len(orbitals.occupied) == num_occ
    assert len(orbitals.virtual) == num_unocc
    
    # Check energies.
    assert orbitals.HOMO_energy == pytest.approx(homo)
    assert orbitals.LUMO_energy == pytest.approx(lumo)
    assert orbitals.HOMO_LUMO_energy == pytest.approx(lumo - homo)
    
    # Check ordering.
    assert [orbital.level for orbital in orbitals] == list(range(1, num_occ + num_unocc +1))


def check_rendered_image(base_name):
    """Check whether the eight separate images of a given rendered image have been created."""
    for extension in ["jpg", "png"]:
        for orientation in ["x0y0z0", "x0y90z0", "x45y45z45", "x90y0z0"]:
            assert Path(base_name.parent, f"{base_name.name}.{orientation}.{extension}").exists()


def check_report(report_folder, base_name, *,
    unrestricted = False,
    optimisation = False,
    dipole = True,
    vibrations = False,
    excited_states = False,
    absorption = False,
    nmr = False,
    s1_tdm = False,
    s1_diff = False,
    t1_diff = False,
    s1_nto = False,
    t1_nto = False,
    vertical_emission = False,
    adiabatic_emission = False):
    """
    Check that certain report and image files have been correctly rendered.
    """
    # First, check the report file itself is there
    assert (Path(report_folder, f"{base_name}.journal.pdf").exists() and Path(report_folder, f"{base_name}.traditional.pdf").exists()) or Path(report_folder, f"{base_name}.pdf").exists()
    
    # Now check for images.
    image_folder = Path(report_folder, "image")
    # HOMO and LUMO
    for image in (["HOMO", "LUMO"] if not unrestricted else ["HOMO (alpha)", "HOMO (beta)", "LUMO (alpha)", "LUMO (beta)"]):
        check_rendered_image(Path(image_folder, f"{image}", f"{base_name}.{image}"))
    
    # HOMO-LUMO overlap
    if not unrestricted:
        check_rendered_image(Path(image_folder, "HOMO LUMO", f"{base_name}.HOMO_LUMO"))
    
    else:
        check_rendered_image(Path(image_folder, "HOMO LUMO", f"{base_name}.alpha_HOMO_LUMO"))
        check_rendered_image(Path(image_folder, "HOMO LUMO", f"{base_name}.beta_HOMO_LUMO"))
    
    # Structure and density.
    check_rendered_image(Path(image_folder, "Structure", f"{base_name}.structure"))
    check_rendered_image(Path(image_folder, "Density", f"{base_name}.SCF"))
    
    # Dipole.
    if dipole:
        check_rendered_image(Path(image_folder, "Dipole Moment", f"{base_name}.dipole"))
    
    if optimisation:
        assert Path(image_folder, f"{base_name}.SCF_graph.png").exists()
    
    # HOMO-LUMO diagrams.
    if not unrestricted:
        assert Path(image_folder, "Orbital Diagram", f"{base_name}.HOMO_LUMO.png").exists()
        assert Path(image_folder, "Orbital Diagram", f"{base_name}.orbitals.png").exists()
    
    else:
        assert Path(image_folder, "Orbital Diagram", f"{base_name}.alpha_HOMO_LUMO.png").exists()
        assert Path(image_folder, "Orbital Diagram", f"{base_name}.beta_HOMO_LUMO.png").exists()
        assert Path(image_folder, "Orbital Diagram", f"{base_name}.alpha_orbitals.png").exists()
        assert Path(image_folder, "Orbital Diagram", f"{base_name}.beta_orbitals.png").exists()
        
    # NMR.
    if nmr:
        assert len(Path(image_folder, "NMR").iterdir()) != 0
    
    # Excited state stuff.
    if s1_tdm:
        check_rendered_image(Path(image_folder, "S(1) Transition Dipole Moment", f"{base_name}.S(1)_dipole"))
    
    if s1_diff:
        check_rendered_image(Path(image_folder, "S(1)", f"{base_name}.S(1)_difference_density"))
    
    if t1_diff:
        check_rendered_image(Path(image_folder, "T(1)", f"{base_name}.T(1)_difference_density"))
        
    if s1_nto:
        check_rendered_image(Path(image_folder, "S(1)", f"{base_name}.S(1)_NTO"))
    
    if t1_nto:
        check_rendered_image(Path(image_folder, "T(1)", f"{base_name}.T(1)_NTO"))
        
    if excited_states:
        assert Path(image_folder, f"{base_name}.excited_states.png").exists()
    
    if absorption:
        assert Path(image_folder, f"{base_name}.simulated_absorption_spectrum.png").exists()
        
    if vertical_emission:
        assert Path(image_folder, f"{base_name}.vertical_S(1)_emission_states.png").exists()
        assert Path(image_folder, f"{base_name}.simulated_vertical_S(1)_emission_spectrum.png").exists()
    
    if adiabatic_emission:
        assert Path(image_folder, f"{base_name}.adiabatic_S(1)_emission_states.png").exists()
        assert Path(image_folder, f"{base_name}.simulated_adiabatic_S(1)_emission_spectrum.png").exists()
    
    # Vibrations.
    if vibrations:
        assert Path(image_folder, f"{base_name}.simulated_frequencies.png").exists()


import scipy.constants
import warnings
import itertools
import math

from digichem.exception.base import Result_unavailable_error


class Result_object():
    """
    Top level class for objects that are designed to hold calculation results.
    """
    
    # A warning issued when attempting to merge non-equivalent objects
    MERGE_WARNING = "Attempting to merge lists of results that are not identical; non-equivalent results will be ignored"
    
    def safe_get(self, *attr_names, default = None):
        """
        Access an attribute of this object, returning default if the attribute is None or is not available (raises Result_unavailable_error).
        """
        # DO YOU LIKE MY B-TEAM SPAGHETTI CODE!?
        
        # Get the first name, which we'll be getattr'ing.
        first_name = attr_names[0]
        remaining_names = attr_names[1:]
        
        # Get.
        try:
            attr = getattr(self, first_name)
            
            if attr is not None and len(remaining_names) > 0:
                if isinstance(attr, Result_object):
                    # We have remaining names to resolve, call attr's safe_get().
                    attr = attr.safe_get(*remaining_names, default = default)
                else:
                    # We have remaining names to resolve, but our attr is not a Result_set (so we can't use safe_get()).
                    for remaining_name in remaining_names:
                        attr = getattr(attr, remaining_name)
        except Result_unavailable_error:
            return default
        
        # And done.
        return attr
    
    @classmethod
    def wavelength_to_energy(self, wavelength):
        """
        Convert a wavelength (in nm) to energy (in eV).
        """
        # e = (c * h) / λ
        return ((scipy.constants.speed_of_light * scipy.constants.Planck) / (wavelength / 1000000000)) / scipy.constants.eV
    
    @classmethod
    def energy_to_wavelength(self, energy):
        """
        Convert an energy (in eV) to wavelength (in nm).
        """
        # λ = (c * h) / e
        return ((scipy.constants.speed_of_light * scipy.constants.Planck) / (energy * scipy.constants.eV)) * 1000000000
    
    @classmethod
    def wavenumbers_to_energy(self, wavenumbers):
        """
        Convert wavenumbers (in cm-1) to energy (in eV).
        """
        return self.wavelength_to_energy((1 / wavenumbers) * 10000000)
    
    @classmethod
    def merged_attr(self, name, objects):
        """
        Merge a number of str like attributes.
        
        :param name: The name of the attribute to merge.
        :param objects: A list of objects to merge from.
        :return: A single string containing the merged attributes.
        """
        attributes = list(dict.fromkeys([getattr(obj, name) for obj in objects]))
        attributes = [attribute for attribute in attributes if attribute is not None]
        return ", ".join(attributes) if len(attributes) > 0 else None
    
    
    @classmethod
    def merge(self, *multiple_objects):
        """
        Merge multiple Result objects of the same type into a single Result object.
        
        Note that this default implementation is intended for objects that cannot actually be merged;
        truly mergable objects should implement their own merge() method.
        """
        # Check all items are the same.
        if not all(obj == multiple_objects[0] for obj in multiple_objects if obj is not None):
            warnings.warn(self.MERGE_WARNING)
            
        return multiple_objects[0]
    
    def calculate(self, item, digichem_options):
        """
        Retrieve/calculate an on-demand value, caching the value if it has not already been retrieved.

        This function is a wrapper around _get_dump_(), caching the calculation result to avoid
        expensive recalculation.

        :param item: The key corresponding to an item in this object's _get_dump_() dict.
        :param digichem_options: Digichem options that will be passed to _get_dump_() (unused on a cache hit).
        """
        if not hasattr(self, "_dump_cache"):
            self._dump_cache = {}

        if item in self._dump_cache:
            # Cache hit.
            return self._dump_cache[item]

        else:
            # Cache miss.
            # Generate (and cache) the data.
            self._dump_cache[item] = self.generate_for_dump()[item](digichem_options)

            # And return of course.
            return self._dump_cache[item]
    
    def generate_for_dump(self):
        """
        Method used to get a dictionary used to generate on-demand values for dumping.
        
        This functionality is useful for hiding expense properties from the normal dump process, while still exposing them when specifically requested.
        
        Each key in the returned dict is the name of a dumpable item, each value is a function to call with digichem_options as its only param.
        """
        warnings.warn("generate_for_dump() is deprecated, use calculate() or _get_dump_() instead", DeprecationWarning)
        return self._get_dump_()
    
    def dump(self, digichem_options, all = False):
        # First, get simple key:value pairs.
        base_dict = self._dump_(digichem_options, all)

        # Then extend with generator ones if we have any.
        if all:
            generators = self._get_dump_()
            for key in generators:
                base_dict[key] = generators[key](digichem_options)
        
        return base_dict
    
    def _dump_(self, digichem_options, all):
        """
        Abstract function that is called to dump the value of the result object to a primitive type, suitable for serializing with yaml.
        
        For real objects, this will normally be a dict with (at least) the following items:
         - 'value': The value of the result (for example, 34.5.
         - 'units': The units of the result (for example, k m^-s).
        """
        raise NotImplementedError("Implement in subclass")
    
    def _get_dump_(self):
        """
        Method used to get a dictionary used to generate on-demand values for dumping.
        
        This functionality is useful for hiding expense properties from the normal dump process, while still exposing them when specifically requested.
        
        Each key in the returned dict is the name of a dumpable item, each value is a function to call with digichem_options as its only param.
        """
        return {}


class Floatable_mixin():
    """
    A mixin class for result objects that have an unambigious numerical representation.
    
    This mixin class expects the implementing class to define __float__().
    """
    
    def __lt__(self, other):
        """
        Is this class less than another?
        """
        return float(self) < float(other)
    
    def __le__(self, other):
        """
        Is this class equal or less than another?
        """
        return math.isclose(float(self), float(other)) or float(self) < float(other)
    
    def __eq__(self, other):
        """
        Is this class equal to another?
        """
        #return float(self) == float(other)
        return math.isclose(float(self), float(other))
    
    def __ge__(self, other):
        """
        Is this class greater or equal to another?
        """
        return math.isclose(float(self), float(other)) or float(self) > float(other)
    
    def __gt__(self, other):
        """
        Is this class greater than another?
        """
        return float(self) > float(other)
    

class Unmergeable_container_mixin():
    """
    Mixin for Result_containers that cannot be merged.
    
    Classes which inherit from this mixin should also set a class attribute with the name MERGE_WARNING, which should be a message to be displayed when attempting to merge this container with another.
    """
    
    @classmethod
    def false_merge(self, *multiple_lists, **kwargs):
        """
        'Merge' a number of containers of the same type.
        
        As Unmergeable_container objects cannot be merged, this method will instead check that all given containers are equivalent, issuing warnings if this is not the case.
        
        :param multiple_lists: A number of container objects to 'merge'.
        :param kwargs: Key-word arguments to pass to the returned container.
        """
        items = self(multiple_lists[0], **kwargs)
        
        # Check all other lists are the same.
        for item_list in multiple_lists[1:]:
            # If this list has atoms and our current doesn't, use this as our base list.
            if len(item_list) > 0 and len(items) == 0:
                items = item_list
            else:
                for index, item in enumerate(item_list):
                    try:
                        other_item = items[index]
                        if not self.are_items_equal(item, other_item):
                            warnings.warn(self.MERGE_WARNING)
                    except IndexError:
                        warnings.warn(self.MERGE_WARNING)
                
        # Return the 'merged' list.
        return items
    
    @classmethod
    def are_items_equal(self, item, other_item):
        """
        A method which determines whether two items are the same for the purposes of merging.
        
        This default implementation simply checks for equivalence between the two items.
        """
        return item == other_item
        
class Result_container(list, Result_object):
    """
    Top level class for Result_objects that hold a list of results.
    """
    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)
        Result_object.__init__(self)
    
    def assign_levels(self):
        """
        (Re)assign total levels of the objects in this list.
        
        The contents of this list will be modified in place.
        """
        total_level = 0
        
        for state in self:            
            # Increment total level.
            total_level += 1
                
            # Set level
            state.level = total_level
            
    @classmethod
    def merge_default(self, *multiple_lists, **kwargs):
        """
        The default implementation of the merge method.
        """
        # First, get a new list containing all objects.
        merged_list = self(itertools.chain(*multiple_lists), **kwargs)
        
        # Now sort.
        # This works because the contained objects are all Floatables (see Floatable_mixin) and therefore have well-defined sort mechanics,
        # or otherwise define their own __eq__ and related methods.
        merged_list.sort()
        
        # Next, we need to reassign levels of the contained objects.
        merged_list.assign_levels()
        
        # All done.
        return merged_list
            
    @classmethod
    def merge(self, *multiple_lists, **kwargs):
        """
        Merge multiple lists of of the same type into a single, ordered list.
        
        Note that this method will return a new list object, but will modify the contained objects (eg, object.level) in place.
        Inheriting classes may safely override this method.
        """
        if hasattr(self, "false_merge"):
            return self.false_merge(*multiple_lists, **kwargs)
        else:
            return self.merge_default(*multiple_lists, **kwargs)
        
    def _dump_(self, digichem_options, all):
        return [item.dump(digichem_options, all) for item in self]
    


# Miscellaneous functions for handling date and time.

# General imports.
import math

def latest_datetime(*datetimes):
    """
    Return the latest of an iterable of datetimes.
    
    :param datetimes:  Datetimes, of which the latest will be returned.
    :returns: The latest of the given datetimes, or None if no datetimes were given.
    """
    try:
        datetimes = list(datetimes)
        datetimes.sort()
        return datetimes[-1]
    except IndexError:
        # No datetimes.
        return None

def total_timedelta(*timedeltas):
    """
    Return the sum of multiple timedeltas.
    
    :param timedeltas: Timedelta objects to be summed.
    :returns: A new timedelta object that is the sum of the objects given, or None if no timedeltas were given.
    """
    try:
        return sum(timedeltas[1:], timedeltas[0])
    except IndexError:
        return None


def date_to_string(datetime_object):
    """
    Convert a datetime object to a standard string representation.
    
    :param datetime_object: The date and time to convert
    :return: A string representation of datetime_object.
    """
    return datetime_object.strftime("%d/%m/%Y %H:%M:%S")

def timedelta_to_string(timedelta_object):
    """
    Convert a timedelta object to a standard string representation.
    
    :param timedelta_object: The time difference to convert
    :return: A string representation of timedelta_object.
    """
    hours = math.floor(timedelta_object.seconds / 3600)
    minutes = math.floor((timedelta_object.seconds - hours * 3600) / 60)
    # Seconds lacks microsecond accuracy, but we can add it on.
    seconds = round((timedelta_object.seconds + timedelta_object.microseconds /1000000) - (hours * 3600 + minutes * 60))
    
    date_str = ""
    add = False
    
    # Only add each section if not zero.
    if timedelta_object.days != 0:
        date_str += "{} d, ".format(timedelta_object.days)
        add = True
        
    if hours != 0 or add:
        date_str += "{} h, ".format(hours)
        add = True
        
    if minutes != 0 or add:
        date_str += "{} m, ".format(minutes)
        add = True
        
    date_str += "{} s".format(seconds)
    
    return date_str


"""Tests for parsing input files"""

import pytest
from pathlib import Path

from digichem.test.util import pyridine_si_v2, pyridine_si_v1, pyridine_cml,\
    result_files, ethane_xyz, benzene_cdx, cyclopentane_com
from digichem.input.digichem_input import si_from_file

@pytest.mark.parametrize("file_path", [
        pyridine_si_v2,
        pyridine_si_v1
     ])
def test_si_reading(file_path):
    """
    Test whether we can correctly read from an existing si file.
    """
    si_file = si_from_file(file_path)
    
    # Check charge and mult.
    assert si_file.charge == 0
    assert si_file.multiplicity == 1
    
    # Check geometry.
    # All files should contain data for pyridine.
    assert len(si_file.atoms) == 11
    assert str(si_file.formula) == "NC5H5"


def test_si_writing(tmp_path):
    """
    Test whether we can correctly write a new .si file.
    """
    # Read in an input file.
    si_file = si_from_file(pyridine_cml)
    
    # Write to our temp dir.
    new_si_path = Path(tmp_path, "Pyrdine.si")
    with open(new_si_path, "wt") as out_file:
        si_file.to_file(out_file)
        
    # Parse the newly written file.
    new_si_file = si_from_file(new_si_path)
    
    # Check they're the same.
    assert new_si_file == si_file


@pytest.mark.parametrize("file_path, sha", [
        [result_files['gaussian'][0], "ebd4f4e9f81cdb57ec2d2f2e1fba9cef0698f4c0"],
        [result_files['turbomole'][0], "60a8ebd9e701b849cfccd9cbb41684519a7fdf0b"],
        [result_files['orca'][0], "e48e7f653f4e67c1bd4c5c4bb76405fad2d441d0"],
     ])
def test_si_history(file_path, sha):
    """
    Test whether the history attribute is set properly.
    """
    si_file = si_from_file(file_path)
    
    assert si_file.history == sha
    
    dump = si_file.dict
    
    assert dump['history'] == sha


@pytest.mark.parametrize("file_path, format", [
    (cyclopentane_com, "mol"),
    (ethane_xyz, "cml"),
    (benzene_cdx, "xyz"),
    (result_files['gaussian'][0], "com")
])
def test_si_file_convert(file_path, format, tmp_path):
    """
    Test whether we can convert from arbitrary formats to another format.
    """
    si_file = si_from_file(file_path)
    si_file.to_format(format, (tmp_path / "file").with_suffix("." + format))


from digichem.result.alignment.FAP import Furthest_atom_pair
from statistics import  mean
import math
import statistics as stats
from digichem.result.alignment import Axis_swapper_mix

class Average_angle(Furthest_atom_pair, Axis_swapper_mix):
    """
    An enhancement to the Kebab skewer method for estimating molecule linearity, featuring a different method to determine a molecule's long axis.
    """
    
    # Names that uniquely describe this alignment protocol.
    CLASS_HANDLE = ["Average Angle", "AA", "Kebab+"]
    
    def align_axes(self):
        """
        Realign the axes of our coordinate system, so they have the following meaning:
        
        X-axis: The long axis (the kebab skewer), which we define as passing through the pair of atoms with the greatest separation in our set.
        Y-axis: The middle axis, which we define as perpendicular to the the X-axis (obviously) and passing through the atom that is furthest from the X-axis. Note that the Y-axis only has to pass through one atom; there may not be a corresponding atom on the other side (but there will be if the molecule is symmetrical about the X-axis).
        Z-axis: The short axis, defined as perpendicular to both the X and Y-axes (so we have no choice where this goes).
        
        :return: Nothing. The atoms are rearranged in place.
        """
        self.align_X()
        self.align_Y()
        self.align_Z()
    
    def get_average_angle(self, thetas):
        """
        Get the average angle from a list of angles.
        """
        average_sin = stats.mean([math.sin(angle) for angle in thetas])
        average_cos = stats.mean([math.cos(angle) for angle in thetas])
        
        raw_angle = math.atan(average_sin / average_cos)
        
        # Adjust.
        if average_sin > 0 and average_cos > 0:
            return raw_angle
        elif average_cos < 0:
            return raw_angle + math.pi
        else:
            return raw_angle + math.pi * 2
        
    
    def align_X(self):
        """
        Align the X axis.
        
        You do not need to call this method yourself; it is called as part of align_axes().
        
        Unlike normal Kebab, we don't align our X axis along the line drawn between our two most separated atoms. Instead, our X axis is defined so as to pass as close to as many points as possible, more accurately passing through the molecule.
        """
        # First get a list of all our coordinates.
        coords = self.get_coordinate_list()
        
        # Set our origin so as to be in the middle of all our points.
        self.translate((-mean(coords[0]) , -mean(coords[1]), -mean(coords[2])))
        
        # Get an average angle from all our points.
        # Rotate xy coords (along Z axis).
        
        theta = self.get_average_angle(([math.atan2(atom.coords[1], atom.coords[0]) for atom in self]))
        self.rotate_XY(theta)
        
        # Rotate xz coords (along Y axis).
        # Get our angle again.
        
        theta = self.get_average_angle([math.atan2(atom.coords[2], atom.coords[0]) for atom in self])
        self.rotate_XZ(theta)
        
    def align_Y(self):
        """
        Align the Y axis.
        
        You do not need to call this method yourself; it is called as part of align_axes().
        """
        # Rotate xz coords (along Y axis).
        # Get our angle.
        theta = self.get_average_angle([math.atan2(atom.coords[2], atom.coords[1]) for atom in self])
        self.rotate_YZ(theta)
        
    def align_Z(self):
        """
        Align the Z axis.
        
        You do not need to call this method yourself; it is called as part of align_axes().
        
        In kebab+, this method can reorientate the molecule, as it ensures X is the longest axis, Y is the second longest, and Z is the shortest.
        """
        self.reassign_axes()
        
#         # If Z is bigger than Y, rotate.
#         if (self.Z_length > self.Y_length):
#             self.rotate_YZ(math.pi/2)
#         
#         # If Y is bigger than X, rotate.
#         if (self.Y_length > self.X_length):
#             self.rotate_XY(math.pi/2)
#         
#         return
        
        
        
        
        
#         # Get the length of each axis.
#         axes = [(0, self.get_axis_length(0)), (1, self.get_axis_length(1)), (2, self.get_axis_length(2))]
#         # Sort them in terms of length.
#         axes.sort(key = lambda axis: axis[1], reverse = True)
#         new_axes = (axes[0][0], axes[1][0], axes[2][0])
#         # Rearrange axes.
#         self.swap_axes(new_axes)


# General imports.
from pathlib import Path
import pwd
import os

from digichem.exception.base import Digichem_exception
from digichem.result.orbital import Molecular_orbital_list,\
    Beta_orbital
from digichem.result.metadata import Metadata
from digichem.result.result import Result_set
from digichem.result.atom import Atom_list
from digichem.result.tdm import Transition_dipole_moment
from digichem.result.excited_state import Excited_state_list
from digichem.result.energy import Energies
from digichem.result.ground_state import Ground_state
from digichem.result.soc import SOC_list
from digichem.result.dipole_moment import Dipole_moment
from digichem.result.vibration import Vibrations_list
from digichem.result.emission import Relaxed_excited_state
from digichem.result.nmr import NMR_shielding, NMR_spin_couplings_list, NMR_list
from digichem.result.alignment.base import Minimal, Alignment


# NOTE: This is a repeat of the list in util to avoid circular import nonsense.
custom_parsing_formats = [
    "sir",
    "sij",
    "yaml",
    "json"
]

class Parser_abc():
    """ABC for all parsers."""

    def __init__(self, *, raw_data = None, options, **kwargs):
        """
        Top level constructor for calculation parsers.
        """
        # An object that we will populate with raw results.
        self.data = raw_data
        
        # A result set object that we'll populate with results.
        self.results = None

        # Config options.
        self.options = options
        
        # Parse (if we haven't already).
        try:
            if self.data is None:
                self.parse()
        except Exception:
            raise Digichem_exception("Error parsing calculation result '{}'".format(self.description))
    
    @property
    def name(self):
        """
        Short name to describe this calculation result.
        """
        return "Parser"
    
    @property
    def description(self):
        """
        A name/path that describes the file(s) being parsed, used for error messages etc.
        """
        return "Parser"

    def parse(self):
        """
        Extract results from our output files.
        """
        self._parse()
    
        # Make any final adjustments.
        self.post_parse()
    
    def _parse(self):
        """
        Extract results from our output files.
        """
        raise NotImplementedError("Implement in subclass")
        
    @classmethod
    def get_current_username(self):
        """
        Get the username of the currently logged in user.
        
        :returns: The username as a string, or None if the username could not be determined.
        """
        # Based on https://stackoverflow.com/questions/842059/is-there-a-portable-way-to-get-the-current-username-in-python
        try:
            return pwd.getpwuid(os.getuid()).pw_name
        
        except Exception as e:
            return None
    
    def post_parse(self):
        """
        Perform any required operations after line-by-line parsing.
        """ 
        # Add current username.
        # TODO: It would probably be better if we used the name of the user who owns the output file, rather than the current user...
        self.data.metadata['user'] = self.get_current_username()
            
    def process_all(self):
        """
        Get all the Result set objects produced by this parser.
        
        :param options: A Digichem options nested dictionary containing options to control parsing.
        :return: A list of the populated result sets.
        """
        self.process()
        return [self.results]
    
    def process(self):
        """
        Get a Result set object from this parser.
        
        :param options: A Digichem options nested dictionary containing options to control parsing.
        :return: The populated result set.
        """
        # Get our result set.
        self.results = Result_set(
            _id = self.data._id,
            metadata = Metadata.from_parser(self),
            aux = self.data._aux if hasattr(self.data, '_aux') else None
            )
        
        alignment_class = Alignment.from_class_handle(self.options['alignment']) if self.options['alignment'] is not None else Minimal
        
        # First get our list of MOs (because we need them for excited states too.)
        self.results.orbitals = Molecular_orbital_list.from_parser(self)
        self.results.beta_orbitals = Molecular_orbital_list.from_parser(self, cls = Beta_orbital)
        
        # Assign total levels.
        self.results.orbitals.assign_total_level(self.results.beta_orbitals)
        
        # Our alignment orientation data.
        self.results.atoms = alignment_class.from_parser(self)
        self.results.raw_atoms = Atom_list.from_parser(self)
        
        # TEDM and TMDM.
        self.results.transition_dipole_moments = Transition_dipole_moment.from_parser(self)
        
        # Excited states.
        self.results.excited_states = Excited_state_list.from_parser(self)
        
        # Energies.
        self.results.energies = Energies.from_parser(self)
        
        # Our ground state.
        self.results.ground_state = Ground_state.from_parser(self)
        
        # And a similar list but also including the ground.
        self.results.energy_states = Excited_state_list()
        self.results.energy_states.append(self.results.ground_state)
        self.results.energy_states.extend(self.results.excited_states)
        
        # SOC.
        self.results.soc = SOC_list.from_parser(self)
        
        # PDM
        self.results.pdm = Dipole_moment.from_parser(self)
        
        # Frequencies.
        self.results.vibrations = Vibrations_list.from_parser(self)
        
        # NMR.
        self.results.nmr_shieldings = NMR_shielding.dict_from_parser(self)
        self.results.nmr_couplings = NMR_spin_couplings_list.from_parser(self)
        self.results.nmr = NMR_list.from_parser(self)
        
        # Finally, try and set emission.
        self.results.emission.vertical, self.results.emission.adiabatic = Relaxed_excited_state.guess_from_results(self.results)
        
        # Return the populated result set for convenience.
        return self.results


class File_parser_abc(Parser_abc):
    """ABC for all parsers."""
    
    def __init__(self, *log_files, raw_data = None, **kwargs):
        """
        Top level constructor for calculation parsers.
        
        :param log_files: A list of output file to analyse/parse. The first log_file given will be used for naming purposes.
        """
        # Set our name.
        self.log_file_paths = self.sort_log_files([Path(log_file) for log_file in log_files if log_file is not None])
        
        # Panic if we have no logs.
        if len(self.log_file_paths) == 0:
            raise Digichem_exception("Cannot parse calculation output; no available log files. Are you sure the given path is a log file or directory containing log files?")
        
        super().__init__(raw_data=raw_data, **kwargs)

    @classmethod
    def from_logs(self, *log_files, **kwargs):
        """
        Intelligent constructor that will attempt to guess the location of aux files from a given log file(s).
        
        :param log_files: Output file(s) to parse or a directory of output files to parse.
        """
        # This default implementation does nothing smart.
        return self(*log_files, **kwargs)
    
    @classmethod
    def sort_log_files(self, log_files):
        """
        Sort a list of log files into a particular order, if required for this parser.
        """
        return log_files
    
    @property
    def log_file_path(self):
        """
        The main log file.
        """
        for log_file in self.log_file_paths:
            if log_file.suffix.lower() == ".log":
                return log_file
            
        return self.log_file_paths[0]
            
    @property
    def name(self):
        """
        Short name to describe this calculation result.
        """
        return self.log_file_path.with_suffix("").name
    
    @property
    def description(self):
        """
        A name/path that describes the file(s) being parsed, used for error messages etc.
        """
        return self.log_file_path
        
    def post_parse(self):
        """
        Perform any required operations after line-by-line parsing.
        """
        super().post_parse()

        # Add our name.
        if self.name is not None:
            self.data.metadata['name'] = self.name
        
        # Set our file paths.
        self.data.metadata['log_files'] = self.log_file_paths
        self.data.metadata['auxiliary_files'] = self.auxiliary_files



import signal

class Uncatchable_exception(BaseException):
    """
    Superclass for exceptions that are not normally caught.
    
    Note that no attempt is made to stop you catching these exceptions if you want, but they will not be caught by 'except Exception' clauses.
    """
    
class Submission_paused(Uncatchable_exception):
    """
    Exception raised during some submission routines.
    
    This exception signals the program that nothing further is to be done from this process; submission will continue in a second process (possible on a different machine on a different planet).
    
    You do not normally want to catch this exception; it is raised during the normal submission process.
    """
    
class Signal_caught(Uncatchable_exception):
    """
    Exception raised when a kill-type signal is sent and caught by the process.
    
    Similarly to KeyboardInterrupt; this exception is not normally caught (so that we exit NOW, this is important because some methods (SLURM, for example) time how long it takes us to shut down, and will send SIGKILL if we're too slow (which is uncatchable by design).
    However, it is valid (and expected) that vital cleanup handlers catch this exception, so long as they re-raise it once done.
    """
    
    def __init__(self, signalnum, stack_frame):
        """
        Constructor for Signal_caught exceptions.
        
        :param signalnum: The signal that was raised.
        :param stackframe: The current stack frame where the signal was caught (can be None).
        """
        self.signalnum = signalnum
        self.stack_frame = stack_frame
        
    @property
    def signal_name(self):
        """
        Get the string name (SIGINT etc) of the signal we caught.
        """
        return self.signal_to_name(self.signalnum)
        
    @classmethod
    def signal_to_name(self, signalnum):
        """
        Convert a singal number to a string (unrecognised signals will return "???"
        """
        try:
            return signal.Signals(signalnum).name
        except Exception:
            return "???"
        
    def __str__(self, *args, **kwargs):
        return "Received signal {} ({})".format(self.signalnum, self.signal_name)
        
        
    @classmethod
    def raise_from_signal(self, signalnum, stack_frame):
        """
        Method designed to be called by a signal handler; automatically raises a Signal_caught exception.
        """
        raise self(signalnum, stack_frame)


# Import top-level object for easier access.
from .base import Result_object
from .base import Result_container
from .base import Unmergeable_container_mixin
from .base import Floatable_mixin
from .result import Result_set


"""Parser(s) for reading from dumped digichem results sets"""
import yaml
import json

from digichem.parse.base import File_parser_abc
from digichem.result.tdm import Transition_dipole_moment
from digichem.result.result import Result_set
from digichem.result.metadata import Metadata
from digichem.result.orbital import Molecular_orbital_list
from digichem.result.atom import Atom_list
from digichem.result.excited_state import Excited_state_list
from digichem.result.energy import Energies
from digichem.result.ground_state import Ground_state
from digichem.result.soc import SOC_list
from digichem.result.dipole_moment import Dipole_moment
from digichem.result.vibration import Vibrations_list
from digichem.result.emission import Relaxed_excited_state
import digichem.log
from digichem.result.alignment.base import Alignment, Minimal
from digichem.result.nmr import NMR_list


class Dump_multi_parser_abc(File_parser_abc):
    """
    ABC for classes that can read multiple result sets from dumped data.
    """
    
    def __init__(self, input_file, *other_log_files, raw_data = None, options , **auxiliary_files):
        """
        Top level constructor for calculation parsers.
        
        :param input_file: The path to read from.
        """
        # Neither other_log_files nor auxiliary_files are currently used...
        super().__init__(input_file, raw_data = raw_data, options = options)
        self.all_results = []
        
    @classmethod
    def from_data(self, input_file, data, options):
        return self(input_file, raw_data = data, options = options)
            
    @property
    def results(self):
        return self.all_results[0]
    
    @results.setter
    def results(self, value):
        pass
    
    def post_parse(self):
        """
        Perform any required operations after line-by-line parsing.
        """
        for result in self.data:
            # Add current username.
            # TODO: It would probably be better if we used the name of the user who owns the output file, rather than the current user...
            result['metadata']['user'] = self.get_current_username()
            
            # Set our file paths.
            result['metadata']['log_files'] = self.log_file_paths
            result['metadata']['auxiliary_files'] = {}
    
    def get_sub_parser(self):
        raise NotImplementedError("Implement in subclass")
            
    def process_all(self):
        """
        Get all the Result set objects produced by this parser.
        
        :return: A list of the populated result sets.
        """
        # Unlike most other parsers, our data can actually contain lots of results.
        self.all_results = [self.get_sub_parser()(self.log_file_path, raw_data = data, options = self.options).process() for data in self.data]
        
        return self.all_results
    
    def process(self):
        """
        Get a Result set object from this parser.
        
        :return: The populated result set.
        """
        self.process_all()
        return self.results
    
    
class Yaml_multi_parser(Dump_multi_parser_abc):
    """
    Parser for reading from dumped result sets in yaml format.
    """
    
    def get_sub_parser(self):
        return Yaml_parser
    
    def _parse(self):
        """
        Extract results from our output files.
        """
        # Read that file.
        digichem.log.get_logger().info("Parsing calculation result '{}'".format(self.description))
        with open(self.log_file_path, "rt") as yaml_file:
            self.data = list(yaml.safe_load_all(yaml_file))
    

class Json_multi_parser(Dump_multi_parser_abc):
    """
    Parser for reading from dumped result sets in JSON format.
    """
    
    def get_sub_parser(self):
        return Json_parser
    
    def _parse(self):
        """
        Extract results from our output files.
        """
        # Read that file.
        digichem.log.get_logger().info("Parsing calculation result '{}'".format(self.description))
        with open(self.log_file_path, "rt") as json_file:
            self.data = json.load(json_file)


class Dump_parser_abc(File_parser_abc):
    """
    ABC for parsers that read dumped data.
    """
    
    def __init__(self, input_file, *other_log_files, raw_data = None, options, **auxiliary_files):
        """
        Top level constructor for calculation parsers.
        
        :param input_file: The path to the input file to read.
        """
        # Neither other_log_files nor auxiliary_files are currently used...
        super().__init__(input_file, raw_data = raw_data, options = options)
        
    @classmethod
    def from_data(self, input_file, data, options):
        return self(input_file, raw_data = data, options = options)
    
    def process_all(self):
        """
        Get all the Result set objects produced by this parser.
        
        :return: A list of the populated result sets.
        """
        self.process()
        return [self.results]
            
    def process(self):
        """
        Get a Result set object from this parser.
        
        :return: The populated result set.
        """
        
        # Get our result set.
        self.results = Result_set(
            _id = self.data.get("_id"),
            metadata = Metadata.from_dump(self.data['metadata'], self.results, self.options)
        )
        
        # First get our list of MOs (because we need them for excited states too.)
        self.results.orbitals = Molecular_orbital_list.from_dump(self.data['orbitals'], self.results, self.options)
        self.results.beta_orbitals = Molecular_orbital_list.from_dump(self.data['beta_orbitals'], self.results, self.options)
        
        alignment_class = Alignment.from_class_handle(self.options['alignment']) if self.options['alignment'] is not None else Minimal
        
        # Our alignment orientation data.
        # The constructor for each alignment class automatically performs realignment.
        self.results.atoms = alignment_class.from_dump(self.data['atoms'], self.results, self.options)
        self.results.raw_atoms = Atom_list.from_dump(self.data['raw_atoms'], self.results, self.options)
        
        # TEDM and TMDM.
        self.results.transition_dipole_moments = Transition_dipole_moment.list_from_dump(self.data['excited_states']['values'], self.results, self.options)
        
        # Excited states.
        self.results.excited_states = Excited_state_list.from_dump(self.data['excited_states'], self.results, self.options)
        
        # Energies.
        self.results.energies = Energies.from_dump(self.data['energies'], self.results, self.options)
        
        # Our ground state.
        self.results.ground_state = Ground_state.from_dump(self.data['ground_state'], self.results, self.options)
        
        # And a similar list but also including the ground.
        self.results.energy_states = Excited_state_list()
        self.results.energy_states.append(self.results.ground_state)
        self.results.energy_states.extend(self.results.excited_states)
        
        # SOC.
        self.results.soc = SOC_list.from_dump(self.data['soc'], self.results, self.options)
        
        # PDM
        self.results.pdm = Dipole_moment.from_dump(self.data['pdm'], self.results, self.options)
        
        # Frequencies.
        self.results.vibrations = Vibrations_list.from_dump(self.data['vibrations'], self.results, self.options)
        
        # NMR.
        if "nmr" in self.data:
            self.results.nmr = NMR_list.from_dump(self.data['nmr'], self.results, self.options)
        
        else:
            self.results.nmr = NMR_list(atoms = self.results.atoms, options = self.options)
            
        # Finally, try and set emission.
        try:
            self.results.emission = self.results.emission.from_dump(self.data['emission'], self.results, self.options)
        
        except Exception:
            digichem.log.get_logger().warning("Failed to parse emission data", exc_info = True)
        
        # If we don't have explicit emission set, try and guess.
        #if len(self.results.emission.vertical) == 0 and len(self.results.emission.adiabatic) == 0:
        #    self.results.emission.vertical, self.results.emission.adiabatic = Relaxed_excited_state.guess_from_results(self.results)
        
        # Return the populated result set for convenience.
        return self.results


class Yaml_parser(Dump_parser_abc):
    """
    Parser for reading from dumped result sets in yaml format.
    """
    
    def _parse(self):
        """
        Extract results from our output files.
        """
        # Read that file.
        digichem.log.get_logger().info("Parsing calculation result '{}'".format(self.description))
        with open(self.log_file_paths[0], "rt") as yaml_file:
            self.data = yaml.safe_load(yaml_file)
            

class Json_parser(Dump_parser_abc):
    """
    Parser for reading from TinyDB json files.
    """
    
    def _parse(self):
        """
        Extract results from our output files.
        """
        raise NotImplementedError("Databases contain many results so it does not make sense to return only one. Try Tinydb_multi_parser instead.")
                
        
    



from collections import OrderedDict

class Layered_dict(list):
    """
    An enhanced OrderedDict object that remembers layers/groups of positions.
    
    A layered dict is essentially a wrapper around a list of ordered dicts. They expose methods found both in list and in dict. Most methods are similar to those found in list (including setting an index), but iteration operates over each dict in turn.
    
    New dicts can either be updated to the last OrderedDict (the same as updating a normal OrderedDict object).
    Or appended to the end of the list, creating a new layer/group.
    
    Layered_dicts are useful for preserving order in cases where you might not know the dict keys in some instances of a layer. Layered_dicts can later be normalised; comparing the order of all keys from all layers (and intelligently skipping layers that contain None) to produce a normal OrderedDict.
    
    Layered_dicts can also be nested to support even more complex cases.
    
    Note that Layered_dict extends from list (and not from dict, at least not yet), be careful if type checking (isinstance(my_layered_dict, dict) == False ...).
    """
    
    def __init__(self, *iterables, **key_values):
        """
        Constructor for Layered_dict objects.
        
        Layered_dicts can be initialised from a single iterable in the same manor as normal dicts. Multiple iterables can also be given as multiple positional args; each iterable will be stored as its own layer. Alternatively, key/value pairs can be given directly as **keyword args (but note that this form is incompatible with the positional args form because the ordering of the layers would be ambiguous.
        
        If no args are given, an empty Layered_dict is created.
        
        :param *iterables: An optional number of iterables to add as new layers.
        :param **key_values: Alternatively, an optional number of keyword arguments which are taken to be key=value pairs to construct a single layer of dict.
        """
        # Init our listy self.
        super().__init__()
        
        if len(iterables) > 0 and len(key_values) > 0:
            raise ValueError("The positional and keyword argument forms are mutually exclusive")
        elif len(iterables) > 0:
            self.extend(iterables)
        elif len(key_values) > 0:
            self.append(key_values)
        else:
            # We'll always have at least one OrderedDict object.    
            self.append({})
    
    @property
    def active_dict(self):
        """
        Return the current active dict object.
        
        The active dict is merely the last dict appended, so use len() to get its index. This method exists only for convenience. 
        """
        return self[-1]
    
    def __setitem__(self, key, value):
        """
        Set the value of one of the layers in this Layered_dict.
        
        Layered_dicts can only contain OrderedDicts, Layered_dict and None as values. Any other values will first be used to construct a new OrderedDict object.
        """
        # First do some type checking.
        if not isinstance(value, (OrderedDict, Layered_dict, type(None))):
            value = OrderedDict(value)
        
        # Now insert.
        list.__setitem__(self, key, value)
        
        # If we inserted None, insert a new LayeredDict too.
        if value is None:
            self.append(OrderedDict)
        
    def __delitem__(self, *args, **kwargs):
        """
        Delete one of the layers of this Layered_dict.
        
        Note that it is impossible to delete all layers; a new empty OrderedDict layer will be created if no layers remain.
        """
        list.__delitem__(self, *args, **kwargs)
        if len(self) == 0:
            self[0] = OrderedDict()
            
    def get(self, layer_index, key, key_index = None):
        """
        Get a value from this Layered_dict.
        
        :raises IndexError: If the one of the given layer indices is not contained in this Layer_dict.
        :raises KeyError: If the given key is not found in the dict at the specified level.
        :param layer_index: A tuple of indices indicating the layer to get from. See items() for further explanation.
        :param key: The name of the key to get at the indicated layer.
        :param key_index: An optional constraint; if given, a KeyError exception is raised if the requested key is not at the the index given by key_index.
        """
        # If layer_index contains multiple parts, we can recurse.
        if len(layer_index) > 1:
            index = layer_index[0]
            if isinstance(self[index], type(self)):
                return self[index].get(layer_index[1:], key, key_index = key_index)
            else:
                # We'be been given a multi-depth layer_index, but our object at that position is not another Layered_dict.
                raise IndexError("Bad layer_index, object ({}) at index {} is not of type {}".format(type(self[index]).__name__, index, type(self).__name__))
            
        # We only have one index to worry about.
        index = layer_index[0]
        
        # Get the key from the dict.
        # But first check to make sure our object at index is not another Layered_dict.
        if isinstance(self[index], type(self)):
            # It is a Layered_dict, raise IndexError.
            raise IndexError("Bad layer_index, object ({}) at index {} is of type {}".format(type(self[index]).__name__, index, type(self).__name__))
        
        # Also check to make sure the dict isn't None.
        elif self[index] is None:
            raise KeyError("Cannot retrieve key '{}' at layer_index '{}'; dict is None".format(key, index))
        
        # Now check the given key_index is the same (if a key_index was given).
        if key_index is not None and list(self[index].keys()).index(key) != key_index:
            raise KeyError("Key/index mismatch; the key '{}' is not at requested index '{}'".format(key, key_index))
    
        return self[index][key]
        
        
    def items(self):
        """
        Generator to support iteration over this Layered_dict.
        
        Each iteration over a Layered_dict returns a 3-membered tuple of the form (layer_index, key_name, value) where:
            layer_index is  a tuple of indices of the current layer being iterated through. For example, the second item of the first layer would be returned as (1,). The third item of the second layer would be (1,2). Each item in layer_index is an index to a nested layer, so the length of layer_index is variable and depends on the degree of nesting. A Layered_dict that contains no other Layered_dict objects will have len(layer_index) == 1 for all items.
            key_name is the name of the key. Note that unlike real dicts, the same key can appear multiple times in a Layered_dict.
            value is the value of the current key in the current dict being iterated through.
            
        Although layered dicts can contain None as placeholder values, iteration will skip these automatically.
        """
        for layer_index, dict_layer in enumerate(self):
            # Some layers can be None place holders; skip these.
            if dict_layer is None:
                continue
            
            # Other layers can be nested Layered_dict objects.
            elif isinstance(dict_layer, Layered_dict):
                for sub_path, key, value in dict_layer.items():
                    # Add our child's index hierarchy to ours.
                    index_path = [layer_index]
                    index_path.extend(sub_path)
                    
                    yield (tuple(index_path), key, value)
            
            # The remaining layers are normal dicts.
            else:
                for key in dict_layer:
                    yield ((layer_index,), key, dict_layer[key])
            
        #while True:
        #    # Nothing more to iterate over.
        #    raise StopIteration()
        
    def update(self, *iterables, **key_values):
        """
        Update the current active dict with new values.
        
        The order of the keys is preserved as for normal OrderedDicts (the first insertion order is preserved). 
        
        :param *iterables: An optional number of iterables to update with. Each iterable will be traversed in the order given and used to update the same dictionary.
        :param **key_values: Alternatively, an optional number of keyword arguments which are used to update the active dict directly.
        :return: This Layered_dict object, for convenience.
        """
        if len(iterables) > 0 and len(key_values) > 0:
            raise ValueError("The positional and keyword argument forms are mutually exclusive")
        elif len(iterables) > 0:
            # Positional arg form.
            # Add each iterable in turn.
            for iterable in iterables:
                if isinstance(iterable, type(self)):
                    # The iterable is another Layered_dict, update with flatten().
                    self.active_dict.update(iterable.flatten())
                else:
                    self.active_dict.update(iterable)
        else:
            # Keyword arg form, update from given keywords.
            self.active_dict.update(**key_values)
            
        return self
        
    def flatten(self):
        """
        Flatten this Layered_dict, returning a single OrderedDict object.
        
        The order of keys in the returned dict will honour the ordering of this Layered_dict object.
        If the same key is present more than once, the first instance determines the order the key appears in, but the last determines its value.
        
        Note that unlike iterating directly over this Layered_dict, the same key cannot appear more than once in the returned flat dict.
        """
        # The implementation here is simple, just update a new dict from each dict we contain.
        return OrderedDict({key:value for index, key, value in self.items()})
    
    def normalise(self, mapping):
        """
        Normalise this Layered_dict according to a mapping so that is aligns with a number of other Layered_dict.
        
        A mapping describes the order in which key/values should be aligned. Although the ordering of keys is defined for each individual Layered_dict, this ordering can be ambiguous if two Layered_dicts contain different keys. Similarly, an individual Layered_dict cannot properly align None placeholder keys because it cannot know how many keys are expected. A mapping resolves this ambiguity.
        
        A mapping is a list of tuples of the form (layer_index, key_name), where:
            layer_index has the same meaning as in items().
            #key_index is the index of the key in dict to get. It is required to differentiate between duplicate keys with different orderings.
            key_name is the name of the key to get (same as in items()).
        
        :param mapping: A list of tuples of the form (layer_index, key_name).
        :return: A list of values as dictated by mapping. Each index of the returned list is controlled by the corresponding index of mapping (and so the two lists will be of the same length). If the index in mapping is None or specifies a key not contained in this dict, None is inserted into the list.
        """
        normal_list = []
        
        for mapped_index in mapping:
            # If the map is None, so is the value we insert.
            mapped_value = None
            
            # Try and get a value from ourself.
            if mapped_index is not None:
                try:
                    mapped_value = self.get(mapped_index[0], mapped_index[1])
                except (IndexError, KeyError):
                    mapped_value = None
            
            # Add to the list.
            normal_list.append(mapped_value)
                
        # And return.
        return normal_list
        
    @classmethod
    def common_mapping(self, layered_dicts):
        """
        Generate a mapping from a list of Layered_dict objects, suitable for passing to normalise().
        
        A mapping is a list of tuples of the form (layer_index, key_name), with the same meaning as in normalise(). It is used to align the keys of multiple dicts.
        
        The mapping generated will contain an entry for each unique (layer_index, key_name) in all of the given layered_dicts. The order of the mapping is well defined, respecting the order of layer_index in the first instance, then the order of keys as found in the respective OrderedDict.
        
        The order is poorly defined, however, for non-identical keys at the same layer_index (because the ordering is ambiguous).
        
        :param layered_dicts: An iterable of Layered_dict objects to generate a mapping from.
        :return: A list describing the mapping.
        """
        # To get our mapping, we'll add non duplicate flattened forms of each layered_dict to our list.
        mapping = []
        for layered_dict in layered_dicts:
            # We can discard the value part as we're only interested in the order for now.
            for index, key, value in layered_dict.items():
                # Add this index_key pair to mapping if not already present.
                # In addition to the normal layer_index, we'll also include the key_index (to support multiple identical dicts with different key ordering).
                if (index, key) not in mapping:
                    mapping.append((index, key))
        
        
        # Finally, we partially re-sort in case any layer indices got added in the wrong order.
        # We won't re-order key names though, so these will still respect their relative orders from their OrderedDicts.
        # If two layer index pairs have OrderedDicts with different keys, the order in which these two OrderedDicts will be combined is undefined (but the keys within them will remember their relative order).
        mapping.sort(key = lambda index_key: index_key[0])
        
        # Done.
        return mapping
    
    @classmethod
    def tabulate(self, layered_dicts):
        """
        Tabulate a list of layered_dicts.
        
        tabulate() is a convenience function, first calling common_mapping() on the list of layered_dicts followed by normalise().
        
        :param layered_dicts: A list of layered_dict objects to tabulate. The order of rows will be the same as the order of this list. Alternatively, a number of OrderedDicts can be included in the list, which will automatically be converted to Layered_dicts.
        :return: A tuple of (headers, table), where headers is a list of table headers (derived from the names of the keys in layered_dicts) and table is the tabulated data (each row will be of the same length, possible padded with None).
        """
        # First convert any OrderedDicts
        layered_dicts = [self(current_dict) if isinstance(current_dict, OrderedDict) else current_dict for current_dict in layered_dicts]
        
        # First, get our mapping.
        mapping = self.common_mapping(layered_dicts)
        
        # Now get our list.
        table = [layered_dict.normalise(mapping) for layered_dict in layered_dicts]
        
        # Return both.
        return ([header[1] for header in mapping], table)
        
        
        
        
        
        
        
        